import json
import os

import pytest

from hipscat.catalog.dataset.dataset import Dataset


def test_dataset_init(base_catalog_info):
    dataset = Dataset(base_catalog_info)
    assert dataset.catalog_info == base_catalog_info
    assert dataset.catalog_name == base_catalog_info.catalog_name
    assert not dataset.on_disk
    assert dataset.catalog_path is None
    assert dataset.catalog_base_dir is None


def test_dataset_wrong_catalog_info(base_catalog_info_data):
    with pytest.raises(TypeError, match="catalog_info"):
        Dataset(base_catalog_info_data)


def test_read_from_hipscat(dataset_path, base_catalog_info_file, assert_catalog_info_matches_dict):
    dataset = Dataset.read_from_hipscat(dataset_path)
    assert dataset.on_disk
    assert dataset.catalog_path == dataset_path
    assert str(dataset.catalog_base_dir) == dataset_path
    with open(base_catalog_info_file, "r", encoding="utf-8") as cat_info_file:
        catalog_info_json = json.load(cat_info_file)
        assert_catalog_info_matches_dict(dataset.catalog_info, catalog_info_json)


def test_read_from_missing_folder(tmp_path):
    wrong_path = os.path.join(tmp_path, "wrong")
    with pytest.raises(FileNotFoundError, match="directory"):
        Dataset.read_from_hipscat(wrong_path)


def test_read_from_empty_folder(tmp_path):
    dataset_path = os.path.join(tmp_path, "dat")
    os.makedirs(dataset_path)
    with pytest.raises(FileNotFoundError, match="catalog info"):
        Dataset.read_from_hipscat(dataset_path)



import json
import os

import pandas as pd
import pytest

from hipscat.catalog import CatalogType
from hipscat.catalog.association_catalog.association_catalog import AssociationCatalog
from hipscat.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hipscat.io.file_io import file_io, file_pointer


def test_read_from_file(
    association_catalog_path_cloud, example_cloud_storage_options, association_catalog_join_pixels
):
    catalog = AssociationCatalog.read_from_hipscat(
        association_catalog_path_cloud, storage_options=example_cloud_storage_options
    )
    assert catalog.on_disk
    assert catalog.catalog_path == association_catalog_path_cloud
    assert len(catalog.get_join_pixels()) == 4
    pd.testing.assert_frame_equal(catalog.get_join_pixels(), association_catalog_join_pixels)

    info = catalog.catalog_info
    assert info.primary_catalog == "small_sky"
    assert info.primary_column == "id"
    assert info.join_catalog == "small_sky_order1"
    assert info.join_column == "id"


def test_empty_directory(
    tmp_dir_cloud,
    example_cloud_storage_options,
    association_catalog_info_data,
    association_catalog_join_pixels,
):
    """Test loading empty or incomplete data"""
    empty_path = os.path.join(tmp_dir_cloud, "path", "empty")
    ## Path doesn't exist
    with pytest.raises(FileNotFoundError):
        AssociationCatalog.read_from_hipscat(empty_path, storage_options=example_cloud_storage_options)

    catalog_path = os.path.join(tmp_dir_cloud, "empty")
    file_io.make_directory(catalog_path, storage_options=example_cloud_storage_options, exist_ok=True)

    ## Path exists but there's nothing there
    with pytest.raises(FileNotFoundError):
        AssociationCatalog.read_from_hipscat(catalog_path, storage_options=example_cloud_storage_options)

    ## catalog_info file exists - getting closer
    file_name = os.path.join(catalog_path, "catalog_info.json")
    file_io.write_string_to_file(
        file_pointer=file_name,
        string=json.dumps(association_catalog_info_data),
        storage_options=example_cloud_storage_options,
    )

    with pytest.raises(FileNotFoundError):
        AssociationCatalog.read_from_hipscat(catalog_path, storage_options=example_cloud_storage_options)

    ## partition_info file exists - enough to create a catalog
    file_name = os.path.join(catalog_path, "partition_join_info.csv")
    file_io.write_dataframe_to_csv(
        association_catalog_join_pixels, file_pointer=file_name, storage_options=example_cloud_storage_options
    )

    catalog = AssociationCatalog.read_from_hipscat(
        catalog_path, storage_options=example_cloud_storage_options
    )
    assert catalog.catalog_name == association_catalog_info_data["catalog_name"]

    file_io.delete_file(
        os.path.join(catalog_path, "catalog_info.json"), storage_options=example_cloud_storage_options
    )
    file_io.delete_file(
        os.path.join(catalog_path, "partition_join_info.csv"), storage_options=example_cloud_storage_options
    )



"""Test construction (and de-construction) of the hipscat ID"""

import healpy as hp
import numpy as np
import numpy.testing as npt
import pytest

from hipscat.pixel_math.hipscat_id import (
    HIPSCAT_ID_HEALPIX_ORDER,
    compute_hipscat_id,
    healpix_to_hipscat_id,
    hipscat_id_to_healpix,
)


def test_single():
    """Single point. Adheres to specification."""
    result = compute_hipscat_id([5], [5])
    expected = [5482513871577022464]

    npt.assert_array_equal(result, expected)


def test_jagged_list():
    """Arrays of mismatched lengths."""
    with pytest.raises(ValueError):
        compute_hipscat_id([5, 1, 5], [5])


def test_short_list():
    """Multiple points that will sit in the same higher-order-pixel.

    There are a handful of points at (5,5), and these should be assigned
    hipscat_id in sequential order, starting at
    5482513871577022464 (0x4C15CD518F800000).

    Interspersed are points at (1,1), which will start at
    5476738131329810432 (0x4C0148503DC00000)
    """
    result = compute_hipscat_id([5, 1, 5], [5, 1, 5])
    expected = [
        5482513871577022464,
        5476738131329810432,  # out of sequence
        5482513871577022465,
    ]
    npt.assert_array_equal(result, expected)


def test_list():
    """Multiple points that will sit in the same higher-order-pixel.

    There are a handful of points at (5,5), and these should be assigned
    hipscat_id in sequential order, starting at
    5482513871577022464 (0x4C15CD518F800000).

    Interspersed are points at (1,1), which will start at
    5476738131329810432 (0x4C0148503DC00000)
    """
    result = compute_hipscat_id([5, 5, 5, 1, 5, 5, 5, 1, 5], [5, 5, 5, 1, 5, 5, 5, 1, 5])
    expected = [
        5482513871577022464,
        5482513871577022465,
        5482513871577022466,
        5476738131329810432,  # out of sequence
        5482513871577022467,
        5482513871577022468,
        5482513871577022469,
        5476738131329810433,  # out of sequence
        5482513871577022470,
    ]
    npt.assert_array_equal(result, expected)


def test_load():
    """Generate a kinda big array and make sure the method completes in under a second.
    If this method is failing due to timeouts, please refactor to keep within the time limit.
    """
    rng = np.random.default_rng(seed=800)
    test_num = 1_000_000

    ra_arr = rng.random(test_num)
    dec_arr = rng.random(test_num)
    result = compute_hipscat_id(ra_arr, dec_arr)

    assert len(result) == test_num


def test_hipscat_id_to_healpix():
    """Test the inverse operation"""
    ids = [
        5482513871577022464,
        5482513871577022465,
        5482513871577022466,
        5476738131329810432,  # out of sequence
        5482513871577022467,
        5482513871577022468,
        5482513871577022469,
        5476738131329810433,  # out of sequence
        5482513871577022470,
    ]

    result = hipscat_id_to_healpix(ids)

    expected = [
        0x1305735463E,
        0x1305735463E,
        0x1305735463E,
        0x130052140F7,
        0x1305735463E,
        0x1305735463E,
        0x1305735463E,
        0x130052140F7,
        0x1305735463E,
    ]

    npt.assert_array_equal(result, expected)


def test_hipscat_id_to_healpix_low_order():
    """Test the inverse operation"""
    ids = [
        5482513871577022464,
        5482513871577022465,
        5482513871577022466,
        5476738131329810432,  # out of sequence
        5482513871577022467,
        5482513871577022468,
        5482513871577022469,
        5476738131329810433,  # out of sequence
        5482513871577022470,
    ]

    result = hipscat_id_to_healpix(ids, target_order=4)

    expected = [1217, 1217, 1217, 1216, 1217, 1217, 1217, 1216, 1217]

    npt.assert_array_equal(result, expected)

def test_healpix_to_hipscat_id_single():
    orders = [3, 3, 4, 1]
    pixels = [0, 12, 1231, 11]
    pixels_at_high_order = [p * (4 ** (HIPSCAT_ID_HEALPIX_ORDER - o)) for o, p in zip(orders, pixels)]
    lon, lat = hp.pix2ang(
        [2**HIPSCAT_ID_HEALPIX_ORDER] * len(orders), pixels_at_high_order, nest=True, lonlat=True
    )
    actual_hipscat_ids = compute_hipscat_id(lon, lat)
    test_hipscat_ids = [healpix_to_hipscat_id(o, p) for o, p in zip(orders, pixels)]
    assert np.all(test_hipscat_ids == actual_hipscat_ids)


def test_healpix_to_hipscat_id_array():
    orders = [3, 3, 4, 1]
    pixels = [0, 12, 1231, 11]
    pixels_at_high_order = [p * (4 ** (HIPSCAT_ID_HEALPIX_ORDER - o)) for o, p in zip(orders, pixels)]
    lon, lat = hp.pix2ang(
        [2**HIPSCAT_ID_HEALPIX_ORDER] * len(orders), pixels_at_high_order, nest=True, lonlat=True
    )
    actual_hipscat_ids = compute_hipscat_id(lon, lat)
    test_hipscat_ids = healpix_to_hipscat_id(orders, pixels)
    assert np.all(test_hipscat_ids == actual_hipscat_ids)


def test_healpix_to_hipscat_id_offset():
    orders = [3, 3, 4, 1]
    pixels = [0, 0, 1231, 11]
    offsets = [0, 1, 0, 0]
    pixels_at_high_order = [p * (4 ** (HIPSCAT_ID_HEALPIX_ORDER - o)) for o, p in zip(orders, pixels)]
    lon, lat = hp.pix2ang(
        [2**HIPSCAT_ID_HEALPIX_ORDER] * len(orders), pixels_at_high_order, nest=True, lonlat=True
    )
    actual_hipscat_ids = compute_hipscat_id(lon, lat)
    test_hipscat_ids = healpix_to_hipscat_id(orders, pixels, offsets)
    assert np.all(test_hipscat_ids == actual_hipscat_ids)



import os
from typing import Any, Dict, List, NewType, Tuple, Union

import fsspec

FilePointer = NewType("FilePointer", str)
"""Unified type for references to files."""


def get_file_protocol(pointer: FilePointer) -> str:
    """Method to parse filepointer for the filesystem protocol.
        if it doesn't follow the pattern of protocol://pathway/to/file, then it
        assumes that it is a localfilesystem.

    Args:
        pointer: filesystem pathway pointer
    """

    if not isinstance(pointer, str):
        pointer = str(pointer)

    protocol = fsspec.utils.get_protocol(pointer)

    return protocol


def get_fs(
    file_pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
) -> Tuple[fsspec.filesystem, FilePointer]:
    """Create the abstract filesystem

    Args:
        file_pointer: filesystem pathway
        storage_options: dictionary that contains abstract filesystem credentials
    Raises:
        ImportError if environment cannot import necessary libraries for
            fsspec filesystems.
    """
    if storage_options is None:
        storage_options = {}
    protocol = get_file_protocol(file_pointer)
    file_pointer = get_file_pointer_for_fs(protocol, file_pointer)

    try:
        file_system = fsspec.filesystem(protocol, **storage_options)
    except ImportError as error:
        raise ImportError from error

    return file_system, file_pointer


def get_file_pointer_for_fs(protocol: str, file_pointer: FilePointer) -> FilePointer:
    """Creates the filepathway from the file_pointer. Will strip the protocol so that
            the file_pointer can be accessed from the filesystem
        abfs filesystems DO NOT require the account_name in the pathway
        s3 filesystems DO require the account_name/container name in the pathway

    Args:
        protocol: str filesytem protocol, file, abfs, or s3
        file_pointer: filesystem pathway

    """
    if not isinstance(file_pointer, str):
        file_pointer = str(file_pointer)

    if protocol == "file":
        # return the entire filepath for local files
        if "file://" in file_pointer:
            split_pointer = file_pointer.split("file://")[1]
        else:
            split_pointer = file_pointer
    else:
        split_pointer = file_pointer.split(f"{protocol}://")[1]

    return FilePointer(split_pointer)


def get_full_file_pointer(incomplete_path: str, protocol_path: str) -> FilePointer:
    """Rebuilds the file_pointer with the protocol and account name if required"""
    protocol = get_file_protocol(protocol_path)
    return FilePointer(f"{protocol}://{incomplete_path}")


def get_file_pointer_from_path(path: str, include_protocol: str = None) -> FilePointer:
    """Returns a file pointer from a path string"""
    if include_protocol:
        path = get_full_file_pointer(path, include_protocol)
    return FilePointer(path)


def get_basename_from_filepointer(pointer: FilePointer) -> str:
    """Returns the base name of a regular file. May return empty string if the file is a directory.

    Args:
        pointer: `FilePointer` object to find a basename within

    Returns:
        string representation of the basename of a file.
    """
    return os.path.basename(pointer)


def strip_leading_slash_for_pyarrow(pointer: FilePointer, protocol: str) -> FilePointer:
    """Strips the leading slash for pyarrow read/write functions.
    This is required for their filesystem abstraction
    """
    if protocol != "file" and str(pointer).startswith("/"):
        pointer = FilePointer(str(pointer).replace("/", "", 1))
    return pointer


def append_paths_to_pointer(pointer: FilePointer, *paths: str) -> FilePointer:
    """Append directories and/or a file name to a specified file pointer.

    Args:
        pointer: `FilePointer` object to add path to
        paths: any number of directory names optionally followed by a file name to append to the
            pointer

    Returns:
        New file pointer to path given by joining given pointer and path names
    """
    return FilePointer(os.path.join(pointer, *paths))


def does_file_or_directory_exist(
    pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
) -> bool:
    """Checks if a file or directory exists for a given file pointer

    Args:
        pointer: File Pointer to check if file or directory exists at
        storage_options: dictionary that contains abstract filesystem credentials

    Returns:
        True if file or directory at `pointer` exists, False if not
    """
    file_system, pointer = get_fs(pointer, storage_options)
    return file_system.exists(pointer)


def is_regular_file(pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None) -> bool:
    """Checks if a regular file (NOT a directory) exists for a given file pointer.

    Args:
        pointer: File Pointer to check if a regular file
        storage_options: dictionary that contains abstract filesystem credentials

    Returns:
        True if regular file at `pointer` exists, False if not or is a directory
    """
    file_system, pointer = get_fs(pointer, storage_options)
    return file_system.isfile(pointer)


def find_files_matching_path(
    pointer: FilePointer, *paths: str, storage_options: Union[Dict[Any, Any], None] = None
) -> List[FilePointer]:
    """Find files or directories matching the provided path parts.

    Args:
        paths: any number of directory names optionally followed by a file name.
            directory or file names may be replaced with `*` as a matcher.
        storage_options: dictionary that contains abstract filesystem credentials
    Returns:
        New file pointers to files found matching the path
    """
    matcher = append_paths_to_pointer(pointer, *paths)
    file_system, pointer = get_fs(pointer, storage_options)
    return [get_file_pointer_from_path(x) for x in file_system.glob(matcher)]


def directory_has_contents(
    pointer: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
) -> bool:
    """Checks if a directory already has some contents (any files or subdirectories)

    Args:
        pointer: File Pointer to check for existing contents
        storage_options: dictionary that contains abstract filesystem credentials

    Returns:
        True if there are any files or subdirectories below this directory.
    """
    return len(find_files_matching_path(pointer, "*", storage_options=storage_options)) > 0


def get_directory_contents(
    pointer: FilePointer, include_protocol=False, storage_options: Union[Dict[Any, Any], None] = None
) -> List[FilePointer]:
    """Finds all files and directories in the specified directory.

    NBL This is not recursive, and will return only the first level of directory contents.

    Args:
        pointer: File Pointer in which to find contents
        include_protocol: boolean on whether or not to include the filesystem protocol in the
            returned directory contents
        storage_options: dictionary that contains abstract filesystem credentials

    Returns:
        New file pointers to files or subdirectories below this directory.
    """
    file_system, file_pointer = get_fs(pointer, storage_options)
    contents = file_system.listdir(file_pointer)
    contents = [FilePointer(x["name"]) for x in contents]

    if len(contents) == 0:
        return []

    contents.sort()
    if include_protocol:
        contents = [get_full_file_pointer(x, protocol_path=pointer) for x in contents]
    return contents



from enum import Enum


class CatalogType(str, Enum):
    """Enum for possible types of catalog"""

    OBJECT = "object"
    SOURCE = "source"
    ASSOCIATION = "association"
    INDEX = "index"
    MARGIN = "margin"

    @classmethod
    def all_types(cls):
        """Fetch a list of all catalog types"""
        return [t.value for t in cls]



import os

import pytest

from hipscat.catalog import PartitionInfo
from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_tree_builder import PixelTreeBuilder


@pytest.fixture
def pixel_trees_dir(test_data_dir):
    return os.path.join(test_data_dir, "pixel_trees")


@pytest.fixture
def pixel_tree_1():
    return PixelTreeBuilder.from_healpix([HealpixPixel(0, 11)])


@pytest.fixture
def pixel_tree_2():
    return PixelTreeBuilder.from_healpix(
        [
            HealpixPixel(0, 10),
            HealpixPixel(1, 33),
            HealpixPixel(1, 35),
            HealpixPixel(1, 44),
            HealpixPixel(1, 45),
            HealpixPixel(1, 46),
            HealpixPixel(2, 128),
            HealpixPixel(2, 130),
            HealpixPixel(2, 131),
        ]
    )


@pytest.fixture
def pixel_tree_3():
    return PixelTreeBuilder.from_healpix(
        [
            HealpixPixel(0, 8),
            HealpixPixel(1, 36),
            HealpixPixel(1, 37),
            HealpixPixel(1, 40),
            HealpixPixel(1, 42),
            HealpixPixel(1, 43),
            HealpixPixel(1, 44),
            HealpixPixel(1, 46),
            HealpixPixel(1, 47),
        ]
    )


@pytest.fixture
def aligned_trees_2_3_inner_path(pixel_trees_dir):
    return os.path.join(pixel_trees_dir, "aligned_2_3_inner.csv")


@pytest.fixture
def aligned_trees_2_3_left_path(pixel_trees_dir):
    return os.path.join(pixel_trees_dir, "aligned_2_3_left.csv")


@pytest.fixture
def aligned_trees_2_3_right_path(pixel_trees_dir):
    return os.path.join(pixel_trees_dir, "aligned_2_3_right.csv")


@pytest.fixture
def aligned_trees_2_3_outer_path(pixel_trees_dir):
    return os.path.join(pixel_trees_dir, "aligned_2_3_outer.csv")


@pytest.fixture
def aligned_trees_2_3_inner(aligned_trees_2_3_inner_path):
    partition_info = PartitionInfo.read_from_file(aligned_trees_2_3_inner_path)
    return PixelTreeBuilder.from_healpix(partition_info.get_healpix_pixels())


@pytest.fixture
def aligned_trees_2_3_left(aligned_trees_2_3_left_path):
    partition_info = PartitionInfo.read_from_file(aligned_trees_2_3_left_path)
    return PixelTreeBuilder.from_healpix(partition_info.get_healpix_pixels())


@pytest.fixture
def aligned_trees_2_3_right(aligned_trees_2_3_right_path):
    partition_info = PartitionInfo.read_from_file(aligned_trees_2_3_right_path)
    return PixelTreeBuilder.from_healpix(partition_info.get_healpix_pixels())


@pytest.fixture
def aligned_trees_2_3_outer(aligned_trees_2_3_outer_path):
    partition_info = PartitionInfo.read_from_file(aligned_trees_2_3_outer_path)
    return PixelTreeBuilder.from_healpix(partition_info.get_healpix_pixels())



import os

import pytest

from hipscat.catalog.dataset.dataset import Dataset
from hipscat.io.file_io import file_io, file_pointer


def test_read_from_hipscat(
    dataset_path_cloud,
    base_catalog_info_file_cloud,
    example_cloud_storage_options,
    assert_catalog_info_matches_dict,
):
    dataset = Dataset.read_from_hipscat(dataset_path_cloud, storage_options=example_cloud_storage_options)
    assert dataset.on_disk
    assert dataset.catalog_path == dataset_path_cloud
    assert str(dataset.catalog_base_dir) == dataset_path_cloud
    catalog_info_json = file_io.load_json_file(
        base_catalog_info_file_cloud, storage_options=example_cloud_storage_options
    )
    assert_catalog_info_matches_dict(dataset.catalog_info, catalog_info_json)


def test_read_from_missing_folder(tmp_dir_cloud, example_cloud_storage_options):
    wrong_path = os.path.join(tmp_dir_cloud, "wrong")
    with pytest.raises(FileNotFoundError, match="directory"):
        Dataset.read_from_hipscat(wrong_path, storage_options=example_cloud_storage_options)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


import os
import sys
from importlib.metadata import version

import autoapi

# Define path to the code to be documented **relative to where conf.py (this file) is kept**
sys.path.insert(0, os.path.abspath("../src/"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "hipscat"
copyright = "2023, LINCC Frameworks"
author = "LINCC Frameworks"
release = version("hipscat")
# for example take major/minor
version = ".".join(release.split(".")[:2])

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.mathjax", "sphinx.ext.napoleon", "sphinx.ext.viewcode"]

extensions.append("autoapi.extension")
extensions.append("nbsphinx")
extensions.append("myst_parser")

templates_path = []
exclude_patterns = ["_build", "**.ipynb_checkpoints"]

# This assumes that sphinx-build is called from the root directory
master_doc = "index"
# Remove 'view source code' from top of page (for html, not python)
html_show_sourcelink = False
# Remove namespaces from class/method signatures
add_module_names = False

autoapi_type = "python"
autoapi_dirs = ["../src"]
autoapi_ignore = ["*/__main__.py", "*/_version.py"]
autoapi_add_toc_tree_entry = False
autoapi_member_order = "bysource"

html_theme = "sphinx_rtd_theme"



from typing import List

import pandas as pd

from hipscat.pixel_math import HealpixInputTypes, HealpixPixel, get_healpix_pixel
from hipscat.pixel_tree.pixel_alignment_types import PixelAlignmentType
from hipscat.pixel_tree.pixel_node_type import PixelNodeType
from hipscat.pixel_tree.pixel_tree import PixelTree
from hipscat.pixel_tree.pixel_tree_builder import PixelTreeBuilder

LEFT_TREE_KEY = "left"
RIGHT_TREE_KEY = "right"


# pylint: disable=R0903
class PixelAlignment:
    """Represents how two pixel trees align with each other, meaning which pixels match
    or overlap between the catalogs, and a new tree with the smallest pixels from each tree

    For more information on the pixel alignment algorithm, view this document:
    https://docs.google.com/document/d/1YFAQsGCgeEyEZ1IRIam9BbWdN5h2onV6ndO3VM-qjn0/edit?usp=sharing

    Attributes:
        pixel_mapping: A dataframe where each row contains a pixel from each tree that match, and
            which pixel in the aligned tree they match with
        pixel_tree: The aligned tree generated by using the smallest pixels in each tree. For
            example, a tree with pixels at order 0, pixel 1, and a tree with order 1, pixel 4,5,6,
            and 7, would result in the smaller order 1 pixels in the aligned tree.
        alignment_type: The type of alignment describing how to handle nodes which exist in one tree
            but not the other. Options are:

                - inner - only use pixels that appear in both catalogs
                - left - use all pixels that appear in the left catalog and any overlapping from the right
                - right - use all pixels that appear in the right catalog and any overlapping from the left
                - outer - use all pixels from both catalogs
    """

    PRIMARY_ORDER_COLUMN_NAME = "primary_Norder"
    PRIMARY_PIXEL_COLUMN_NAME = "primary_Npix"
    JOIN_ORDER_COLUMN_NAME = "join_Norder"
    JOIN_PIXEL_COLUMN_NAME = "join_Npix"
    ALIGNED_ORDER_COLUMN_NAME = "aligned_Norder"
    ALIGNED_PIXEL_COLUMN_NAME = "aligned_Npix"

    def __init__(
        self,
        aligned_tree: PixelTree,
        pixel_mapping: pd.DataFrame,
        alignment_type: PixelAlignmentType,
    ) -> None:
        self.pixel_tree = aligned_tree
        self.pixel_mapping = pixel_mapping
        self.alignment_type = alignment_type


# pylint: disable=R0912
def align_trees(
    left: PixelTree,
    right: PixelTree,
    alignment_type: PixelAlignmentType = PixelAlignmentType.INNER,
) -> PixelAlignment:
    """Generate a `PixelAlignment` object from two pixel trees

    A `PixelAlignment` represents how two pixel trees align with each other, meaning which pixels
    match or overlap between the catalogs, and includes a new tree with the smallest pixels from
    each tree

    For more information on the pixel alignment algorithm, view this document:
    https://docs.google.com/document/d/1YFAQsGCgeEyEZ1IRIam9BbWdN5h2onV6ndO3VM-qjn0/edit?usp=sharing

    Args:
        left: The left tree to align
        right: The right tree to align
        alignment_type: The type of alignment describing how to handle nodes which exist in one tree
            but not the other. Options are:

                - inner - only use pixels that appear in both catalogs
                - left - use all pixels that appear in the left catalog and any overlapping from the right
                - right - use all pixels that appear in the right catalog and any overlapping from the left
                - outer - use all pixels from both catalogs

    Returns:
        The `PixelAlignment` object with the alignment from the two trees
    """

    tree_builder = PixelTreeBuilder()
    pixels_to_search = _get_children_pixels_from_trees([left, right], left.root_pixel.pixel)

    while len(pixels_to_search) > 0:
        search_pixel = pixels_to_search.pop(0)
        if search_pixel in left and search_pixel in right:
            left_node = left[search_pixel]
            right_node = right[search_pixel]
            if left_node.node_type == right_node.node_type:
                if left_node.node_type == PixelNodeType.LEAF:
                    # Matching leaf nodes get added to the aligned tree
                    tree_builder.create_node_and_parent_if_not_exist(search_pixel, PixelNodeType.LEAF)
                else:
                    # For matching inner nodes search into their children to check for alignment
                    pixels_to_search += _get_children_pixels_from_trees([left, right], search_pixel)
            else:
                # Nodes with non-matching types: one must be a leaf and the other an inner node
                if left_node.node_type == PixelNodeType.LEAF:
                    tree_with_leaf_node = LEFT_TREE_KEY
                    inner_node = right_node
                else:
                    tree_with_leaf_node = RIGHT_TREE_KEY
                    inner_node = left_node
                if _should_include_all_pixels_from_tree(tree_with_leaf_node, alignment_type):
                    # If the alignment type means fully covering the tree with the leaf node, then
                    # create a leaf node in the aligned tree and split it to match the partitioning
                    # of the other tree to ensure the node is fully covered
                    tree_builder.create_node_and_parent_if_not_exist(search_pixel, PixelNodeType.LEAF)
                    tree_builder.split_leaf_to_match_partitioning(inner_node)
                else:
                    # Otherwise just add the subtree from the inner node to include all the
                    # overlapping pixels
                    tree_builder.create_node_and_parent_if_not_exist(search_pixel, PixelNodeType.INNER)
                    tree_builder.add_all_descendants_from_node(inner_node)
        elif search_pixel in left and search_pixel not in right:
            # For nodes that only exist in one tree, include them if the alignment type means that
            # tree should have all its nodes included
            if _should_include_all_pixels_from_tree(LEFT_TREE_KEY, alignment_type):
                tree_builder.create_node_and_parent_if_not_exist(search_pixel, left[search_pixel].node_type)
                tree_builder.add_all_descendants_from_node(left[search_pixel])
        elif search_pixel in right and search_pixel not in left:
            if _should_include_all_pixels_from_tree(RIGHT_TREE_KEY, alignment_type):
                tree_builder.create_node_and_parent_if_not_exist(search_pixel, right[search_pixel].node_type)
                tree_builder.add_all_descendants_from_node(right[search_pixel])
    tree = tree_builder.build()
    pixel_mapping = _generate_pixel_mapping_from_tree(left, right, tree)
    return PixelAlignment(tree, pixel_mapping, alignment_type)


def _get_children_pixels_from_trees(trees: List[PixelTree], pixel: HealpixInputTypes) -> List[HealpixPixel]:
    """Returns the combined HEALPix pixels that have child nodes of the given pixel from trees

    This returns a list of HEALPix pixels, not the actual child nodes, and does not contain
    duplicate pixels if a pixel appears in multiple trees.

    Args:
        trees (List[PixelTree]): The list of trees to search for children from
        pixel (HealpixPixel | tuple[int,int]): The pixel to search for children at in all trees

    Returns:
        (List[HealpixPixel]) The list of all HEALPix pixels which have children of the given pixel
        in any of the trees.
    """
    pixel = get_healpix_pixel(pixel)
    pixels_to_add = set()
    for tree in trees:
        if pixel in tree:
            for node in tree[pixel].children:
                pixels_to_add.add(node.pixel)
    return list(pixels_to_add)


def _should_include_all_pixels_from_tree(tree_type: str, alignment_type: PixelAlignmentType) -> bool:
    """If for a given alignment type, the left or right tree should include all pixels or just the
    ones that overlap with the other tree.

    Args:
        tree_type (str): 'left' for the left tree and 'right' for the right tree
        alignment_type (PixelAlignmentType): The type of alignment being performed

    Returns:
        A boolean indicating if the given tree type should include all pixels
    """
    left_add_types = [PixelAlignmentType.OUTER, PixelAlignmentType.LEFT]
    right_add_types = [PixelAlignmentType.OUTER, PixelAlignmentType.RIGHT]
    return (tree_type == LEFT_TREE_KEY and alignment_type in left_add_types) or (
        tree_type == RIGHT_TREE_KEY and alignment_type in right_add_types
    )


def _generate_pixel_mapping_from_tree(left: PixelTree, right: PixelTree, aligned: PixelTree) -> pd.DataFrame:
    """Generates a pixel mapping dataframe from two trees and their aligned tree

    The pixel mapping dataframe contains columns for the order and pixel of overlapping pixels in
    the left, right and aligned trees. The trees are searched through and this table is generated

    Args:
        left (PixelTree): the left tree used to generate the alignment
        right (PixelTree): the right tree used to generate the alignment
        aligned (PixelTree): the aligned tree as a result of aligning the left and right trees

    Returns:
        (pd.DataFrame) The pixel mapping dataframe where each row contains a pixel from the aligned
        tree and the pixels in the left and right tree that overlap with it
    """
    pixel_mapping_dict = {
        PixelAlignment.PRIMARY_ORDER_COLUMN_NAME: [],
        PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME: [],
        PixelAlignment.JOIN_ORDER_COLUMN_NAME: [],
        PixelAlignment.JOIN_PIXEL_COLUMN_NAME: [],
        PixelAlignment.ALIGNED_ORDER_COLUMN_NAME: [],
        PixelAlignment.ALIGNED_PIXEL_COLUMN_NAME: [],
    }
    for leaf_node in aligned.root_pixel.get_all_leaf_descendants():
        left_leaf_nodes = left.get_leaf_nodes_at_healpix_pixel(leaf_node.pixel)
        right_leaf_nodes = right.get_leaf_nodes_at_healpix_pixel(leaf_node.pixel)
        if len(left_leaf_nodes) == 0:
            left_leaf_nodes = [None]
        if len(right_leaf_nodes) == 0:
            right_leaf_nodes = [None]
        for left_node in left_leaf_nodes:
            for right_node in right_leaf_nodes:
                pixel_mapping_dict[PixelAlignment.ALIGNED_ORDER_COLUMN_NAME].append(leaf_node.hp_order)
                pixel_mapping_dict[PixelAlignment.ALIGNED_PIXEL_COLUMN_NAME].append(leaf_node.hp_pixel)
                left_order = left_node.hp_order if left_node is not None else None
                left_pixel = left_node.hp_pixel if left_node is not None else None
                pixel_mapping_dict[PixelAlignment.PRIMARY_ORDER_COLUMN_NAME].append(left_order)
                pixel_mapping_dict[PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME].append(left_pixel)
                right_order = right_node.hp_order if right_node is not None else None
                right_pixel = right_node.hp_pixel if right_node is not None else None
                pixel_mapping_dict[PixelAlignment.JOIN_ORDER_COLUMN_NAME].append(right_order)
                pixel_mapping_dict[PixelAlignment.JOIN_PIXEL_COLUMN_NAME].append(right_pixel)
    pixel_mapping = pd.DataFrame.from_dict(pixel_mapping_dict).astype(pd.Int64Dtype())
    return pixel_mapping



"""Tests of catalog functionality"""

import os

import pytest

from hipscat.catalog import Catalog, CatalogType, PartitionInfo
from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_node_type import PixelNodeType
from hipscat.pixel_tree.pixel_tree_builder import PixelTreeBuilder


def test_catalog_load(catalog_info, catalog_pixels):
    catalog = Catalog(catalog_info, catalog_pixels)
    assert catalog.get_healpix_pixels() == catalog_pixels
    assert catalog.catalog_name == catalog_info.catalog_name

    for hp_pixel in catalog_pixels:
        assert hp_pixel in catalog.pixel_tree
        assert catalog.pixel_tree[hp_pixel].node_type == PixelNodeType.LEAF


def test_catalog_load_wrong_catalog_info(base_catalog_info, catalog_pixels):
    with pytest.raises(TypeError):
        Catalog(base_catalog_info, catalog_pixels)


def test_catalog_wrong_catalog_type(catalog_info, catalog_pixels):
    catalog_info.catalog_type = CatalogType.INDEX
    with pytest.raises(ValueError):
        Catalog(catalog_info, catalog_pixels)


def test_partition_info_pixel_input_types(catalog_info, catalog_pixels):
    partition_info = PartitionInfo.from_healpix(catalog_pixels)
    catalog = Catalog(catalog_info, partition_info)
    assert len(catalog.get_healpix_pixels()) == len(catalog_pixels)
    assert len(catalog.pixel_tree.root_pixel.get_all_leaf_descendants()) == len(catalog_pixels)
    for hp_pixel in catalog_pixels:
        assert hp_pixel in catalog.pixel_tree
        assert catalog.pixel_tree[hp_pixel].node_type == PixelNodeType.LEAF


def test_tree_pixel_input(catalog_info, catalog_pixels):
    tree = PixelTreeBuilder.from_healpix(catalog_pixels)
    catalog = Catalog(catalog_info, tree)
    assert len(catalog.get_healpix_pixels()) == len(catalog_pixels)
    assert len(catalog.pixel_tree.root_pixel.get_all_leaf_descendants()) == len(catalog_pixels)
    for pixel in catalog_pixels:
        assert pixel in catalog.pixel_tree
        assert catalog.pixel_tree[pixel].node_type == PixelNodeType.LEAF


def test_tree_pixel_input_list(catalog_info, catalog_pixels):
    catalog = Catalog(catalog_info, catalog_pixels)
    assert len(catalog.get_healpix_pixels()) == len(catalog_pixels)
    assert len(catalog.pixel_tree.root_pixel.get_all_leaf_descendants()) == len(catalog_pixels)
    for pixel in catalog_pixels:
        assert pixel in catalog.pixel_tree
        assert catalog.pixel_tree[pixel].node_type == PixelNodeType.LEAF


def test_wrong_pixel_input_type(catalog_info):
    with pytest.raises(TypeError):
        Catalog(catalog_info, "test")
    with pytest.raises(TypeError):
        Catalog._get_pixel_tree_from_pixels("test")
    with pytest.raises(TypeError):
        Catalog._get_partition_info_from_pixels("test")


def test_get_pixels_list(catalog_info, catalog_pixels):
    catalog = Catalog(catalog_info, catalog_pixels)
    pixels = catalog.get_healpix_pixels()
    assert pixels == catalog_pixels


def test_load_catalog_small_sky(small_sky_dir):
    """Instantiate a catalog with 1 pixel"""
    cat = Catalog.read_from_hipscat(small_sky_dir)

    assert cat.catalog_name == "small_sky"
    assert len(cat.get_healpix_pixels()) == 1


def test_load_catalog_small_sky_order1(small_sky_order1_dir):
    """Instantiate a catalog with 4 pixels"""
    cat = Catalog.read_from_hipscat(small_sky_order1_dir)

    assert cat.catalog_name == "small_sky_order1"
    assert len(cat.get_healpix_pixels()) == 4


def test_cone_filter(small_sky_order1_catalog):
    filtered_catalog = small_sky_order1_catalog.filter_by_cone(315, -66.443, 0.1)
    filtered_pixels = filtered_catalog.get_healpix_pixels()

    assert len(filtered_pixels) == 1
    assert filtered_pixels == [HealpixPixel(1, 44)]

    assert (1, 44) in filtered_catalog.pixel_tree
    assert len(filtered_catalog.pixel_tree.pixels[1]) == 1
    assert filtered_catalog.catalog_info.total_rows is None


def test_cone_filter_big(small_sky_order1_catalog):
    filtered_catalog = small_sky_order1_catalog.filter_by_cone(315, -66.443, 30)
    assert len(filtered_catalog.get_healpix_pixels()) == 4
    assert (1, 44) in filtered_catalog.pixel_tree
    assert (1, 45) in filtered_catalog.pixel_tree
    assert (1, 46) in filtered_catalog.pixel_tree
    assert (1, 47) in filtered_catalog.pixel_tree


@pytest.mark.timeout(5)
def test_cone_filter_multiple_order(catalog_info):
    catalog_pixel_list = [
        HealpixPixel(6, 30),
        HealpixPixel(7, 124),
        HealpixPixel(7, 5000),
    ]
    catalog = Catalog(catalog_info, catalog_pixel_list)
    filtered_catalog = catalog.filter_by_cone(47.1, 6, 30)
    assert filtered_catalog.get_healpix_pixels() == [HealpixPixel(6, 30), HealpixPixel(7, 124)]


def test_cone_filter_empty(small_sky_order1_catalog):
    filtered_catalog = small_sky_order1_catalog.filter_by_cone(0, 0, 0.1)
    assert len(filtered_catalog.get_healpix_pixels()) == 0
    assert len(filtered_catalog.pixel_tree) == 1


def test_empty_directory(tmp_path):
    """Test loading empty or incomplete data"""
    ## Path doesn't exist
    with pytest.raises(FileNotFoundError):
        Catalog.read_from_hipscat(os.path.join("path", "empty"))

    catalog_path = os.path.join(tmp_path, "empty")
    os.makedirs(catalog_path, exist_ok=True)

    ## Path exists but there's nothing there
    with pytest.raises(FileNotFoundError):
        Catalog.read_from_hipscat(catalog_path)

    ## catalog_info file exists - getting closer
    file_name = os.path.join(catalog_path, "catalog_info.json")
    with open(file_name, "w", encoding="utf-8") as metadata_file:
        metadata_file.write('{"catalog_name":"empty", "catalog_type":"source"}')

    with pytest.raises(FileNotFoundError):
        Catalog.read_from_hipscat(catalog_path)

    ## partition_info file exists - enough to create a catalog
    file_name = os.path.join(catalog_path, "partition_info.csv")
    with open(file_name, "w", encoding="utf-8") as metadata_file:
        metadata_file.write("Norder,Dir,Npix")

    catalog = Catalog.read_from_hipscat(catalog_path)
    assert catalog.catalog_name == "empty"


def test_generate_negative_tree_pixels(small_sky_order1_catalog):
    """Test generate_negative_tree_pixels on a basic catalog."""
    expected_pixels = [
        HealpixPixel(0, 0),
        HealpixPixel(0, 1),
        HealpixPixel(0, 2),
        HealpixPixel(0, 3),
        HealpixPixel(0, 4),
        HealpixPixel(0, 5),
        HealpixPixel(0, 6),
        HealpixPixel(0, 7),
        HealpixPixel(0, 8),
        HealpixPixel(0, 9),
        HealpixPixel(0, 10),
    ]

    negative_tree = small_sky_order1_catalog.generate_negative_tree_pixels()

    assert negative_tree == expected_pixels


def test_generate_negative_tree_pixels_order_0(small_sky_catalog):
    """Test generate_negative_tree_pixels on a catalog with only order 0 pixels."""
    expected_pixels = [
        HealpixPixel(0, 0),
        HealpixPixel(0, 1),
        HealpixPixel(0, 2),
        HealpixPixel(0, 3),
        HealpixPixel(0, 4),
        HealpixPixel(0, 5),
        HealpixPixel(0, 6),
        HealpixPixel(0, 7),
        HealpixPixel(0, 8),
        HealpixPixel(0, 9),
        HealpixPixel(0, 10),
    ]

    negative_tree = small_sky_catalog.generate_negative_tree_pixels()

    assert negative_tree == expected_pixels


def test_generate_negative_tree_pixels_multi_order(small_sky_order1_catalog):
    """Test generate_negative_tree_pixels on a catalog with
    missing pixels on multiple order.
    """
    # remove one of the order 1 pixels from the catalog.
    nodes = small_sky_order1_catalog.pixel_tree.root_pixel.children[0].children
    small_sky_order1_catalog.pixel_tree.root_pixel.children[0].children = nodes[1:]

    expected_pixels = [
        HealpixPixel(0, 0),
        HealpixPixel(0, 1),
        HealpixPixel(0, 2),
        HealpixPixel(0, 3),
        HealpixPixel(0, 4),
        HealpixPixel(0, 5),
        HealpixPixel(0, 6),
        HealpixPixel(0, 7),
        HealpixPixel(0, 8),
        HealpixPixel(0, 9),
        HealpixPixel(0, 10),
        HealpixPixel(1, 44),
    ]

    negative_tree = small_sky_order1_catalog.generate_negative_tree_pixels()

    assert negative_tree == expected_pixels



import dataclasses
from dataclasses import dataclass
from typing import Any, Dict, Union

from typing_extensions import Self

from hipscat.catalog.catalog_type import CatalogType
from hipscat.io import FilePointer, file_io


@dataclass
class BaseCatalogInfo:
    """Container class for catalog metadata"""

    catalog_name: str = ""
    catalog_type: CatalogType = None
    total_rows: int = None

    DEFAULT_TYPE = None
    """The default catalog type for this catalog info type. To be overridden by subclasses.
    If specified, we will use this value when no catalog_type is provided."""

    REQUIRED_TYPE = None
    """The required catalog type for this catalog info type. To be overridden by subclasses.
    If specified, the catalog MUST have this type."""

    required_fields = ["catalog_type"]

    def __post_init__(self):
        if not self.catalog_type and self.DEFAULT_TYPE:
            self.catalog_type = self.DEFAULT_TYPE
        elif self.REQUIRED_TYPE and self.catalog_type != self.REQUIRED_TYPE:
            raise ValueError(f"Catalog must have type {self.REQUIRED_TYPE}")
        self._check_required_fields()
        if self.catalog_type not in CatalogType.all_types():
            raise ValueError(f"Unknown catalog type: {self.catalog_type}")

    def __str__(self):
        parameters = dataclasses.asdict(self)
        formatted_string = ""
        for name, value in parameters.items():
            formatted_string += f"  {name} {value}\n"
        return formatted_string

    @classmethod
    def read_from_metadata_file(
        cls, catalog_info_file: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ) -> Self:
        """Read catalog info from the `catalog_info.json` metadata file

        Args:
            catalog_info_file: FilePointer pointing to the `catalog_info.json` file
            storage_options: dictionary that contains abstract filesystem credentials

        Returns:
            A CatalogInfo object with the data from the `catalog_info.json` file
        """
        metadata_keywords = file_io.load_json_file(catalog_info_file, storage_options=storage_options)
        catalog_info_keywords = {}
        for field in dataclasses.fields(cls):
            if field.name in metadata_keywords:
                catalog_info_keywords[field.name] = metadata_keywords[field.name]
        return cls(**catalog_info_keywords)

    def _check_required_fields(self):
        fields_dict = dataclasses.asdict(self)
        for field_name in self.required_fields:
            if field_name not in fields_dict or fields_dict[field_name] is None:
                raise ValueError(f"{field_name} is required in the Catalog Info and a value must be provided")



import os

from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo
from hipscat.io import file_io


def test_read_from_file(
    base_catalog_info_file_cloud, example_cloud_storage_options, assert_catalog_info_matches_dict
):
    base_cat_info_fp = file_io.get_file_pointer_from_path(base_catalog_info_file_cloud)
    catalog_info = BaseCatalogInfo.read_from_metadata_file(
        base_cat_info_fp, storage_options=example_cloud_storage_options
    )
    catalog_info_json = file_io.file_io.load_json_file(
        base_catalog_info_file_cloud, storage_options=example_cloud_storage_options
    )
    assert_catalog_info_matches_dict(catalog_info, catalog_info_json)



from typing import Any, Dict, Tuple, Union

import pandas as pd
from typing_extensions import TypeAlias

from hipscat.catalog.association_catalog.association_catalog_info import AssociationCatalogInfo
from hipscat.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hipscat.catalog.catalog_type import CatalogType
from hipscat.catalog.healpix_dataset.healpix_dataset import HealpixDataset, PixelInputTypes
from hipscat.io import FilePointer, file_io, paths


class AssociationCatalog(HealpixDataset):
    """A HiPSCat Catalog for enabling fast joins between two HiPSCat catalogs

    Catalogs of this type are partitioned based on the partitioning of the left catalog.
    The `partition_join_info` metadata file specifies all pairs of pixels in the Association
    Catalog, corresponding to each pair of partitions in each catalog that contain rows to join.
    """

    # Update CatalogInfoClass, used to check if the catalog_info is the correct type, and
    # set the catalog info to the correct type
    CatalogInfoClass: TypeAlias = AssociationCatalogInfo
    catalog_info: CatalogInfoClass

    JoinPixelInputTypes = Union[list, pd.DataFrame, PartitionJoinInfo]

    def __init__(
        self,
        catalog_info: CatalogInfoClass,
        pixels: PixelInputTypes,
        join_pixels: JoinPixelInputTypes,
        catalog_path=None,
        storage_options: Union[Dict[Any, Any], None] = None,
    ) -> None:
        if not catalog_info.catalog_type == CatalogType.ASSOCIATION:
            raise ValueError("Catalog info `catalog_type` must be 'association'")
        super().__init__(catalog_info, pixels, catalog_path, storage_options=storage_options)
        self.join_info = self._get_partition_join_info_from_pixels(join_pixels)

    def get_join_pixels(self) -> pd.DataFrame:
        """Get join pixels listing all pairs of pixels from left and right catalogs that contain
        matching association rows

        Returns:
            pd.DataFrame with each row being a pair of pixels from the primary and join catalogs
        """
        return self.join_info.data_frame

    @staticmethod
    def _get_partition_join_info_from_pixels(
        join_pixels: JoinPixelInputTypes,
    ) -> PartitionJoinInfo:
        if isinstance(join_pixels, PartitionJoinInfo):
            return join_pixels
        if isinstance(join_pixels, pd.DataFrame):
            return PartitionJoinInfo(join_pixels)
        raise TypeError("join_pixels must be of type PartitionJoinInfo or DataFrame")

    @classmethod
    def _read_args(
        cls, catalog_base_dir: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ) -> Tuple[CatalogInfoClass, PixelInputTypes, JoinPixelInputTypes]:  # type: ignore[override]
        args = super()._read_args(catalog_base_dir, storage_options=storage_options)
        partition_join_info_file = paths.get_partition_join_info_pointer(catalog_base_dir)
        partition_join_info = PartitionJoinInfo.read_from_file(
            partition_join_info_file, storage_options=storage_options
        )
        return args + (partition_join_info,)

    @classmethod
    def _check_files_exist(
        cls, catalog_base_dir: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ):
        super()._check_files_exist(catalog_base_dir, storage_options=storage_options)
        partition_join_info_file = paths.get_partition_join_info_pointer(catalog_base_dir)
        if not file_io.does_file_or_directory_exist(
            partition_join_info_file, storage_options=storage_options
        ):
            raise FileNotFoundError(
                f"No partition join info found where expected: {str(partition_join_info_file)}"
            )



from typing import Any, Dict, Tuple, Union

from typing_extensions import Self, TypeAlias

from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo
from hipscat.io import FilePointer, file_io, paths


class Dataset:
    """A base HiPSCat dataset

    A base dataset contains a catalog_info metadata file and the data contained in parquet files

    TODO - create factory methods to get appropriately-typed datasets for
    some catalog info or catalog directory
    """

    CatalogInfoClass: TypeAlias = BaseCatalogInfo

    def __init__(
        self,
        catalog_info: CatalogInfoClass,
        catalog_path=None,
        storage_options: Union[Dict[Any, Any], None] = None,
    ) -> None:
        """Initializes a Dataset

        Args:
            catalog_info: A catalog_info object with the catalog metadata
            catalog_path: If the catalog is stored on disk, specify the location of the catalog
                Does not load the catalog from this path, only store as metadata
            storage_options: dictionary that contains abstract filesystem credentials
        """
        if not isinstance(catalog_info, self.CatalogInfoClass):
            raise TypeError(f"catalog_info type must be {self.CatalogInfoClass}")

        self.catalog_info = catalog_info
        self.catalog_name = self.catalog_info.catalog_name

        self.catalog_path = catalog_path
        self.on_disk = catalog_path is not None
        self.storage_options = storage_options
        self.catalog_base_dir = file_io.get_file_pointer_from_path(self.catalog_path)

    @classmethod
    def read_from_hipscat(
        cls, catalog_path: str, storage_options: Union[Dict[Any, Any], None] = None
    ) -> Self:
        """Reads a HiPSCat Catalog from a HiPSCat directory

        Args:
            catalog_path: path to the root directory of the catalog
            storage_options: dictionary that contains abstract filesystem credentials

        Returns:
            The initialized catalog object
        """
        catalog_base_dir = file_io.get_file_pointer_from_path(catalog_path)
        cls._check_files_exist(catalog_base_dir, storage_options=storage_options)
        args = cls._read_args(catalog_base_dir, storage_options=storage_options)
        kwargs = cls._read_kwargs(catalog_base_dir, storage_options=storage_options)
        return cls(*args, **kwargs)

    @classmethod
    def _read_args(
        cls, catalog_base_dir: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ) -> Tuple[CatalogInfoClass]:
        catalog_info_file = paths.get_catalog_info_pointer(catalog_base_dir)
        catalog_info = cls.CatalogInfoClass.read_from_metadata_file(
            catalog_info_file, storage_options=storage_options
        )
        return (catalog_info,)

    @classmethod
    def _read_kwargs(
        cls, catalog_base_dir: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ) -> dict:
        return {"catalog_path": str(catalog_base_dir), "storage_options": storage_options}

    @classmethod
    def _check_files_exist(
        cls, catalog_base_dir: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ):
        if not file_io.does_file_or_directory_exist(catalog_base_dir, storage_options=storage_options):
            raise FileNotFoundError(f"No directory exists at {str(catalog_base_dir)}")
        catalog_info_file = paths.get_catalog_info_pointer(catalog_base_dir)
        if not file_io.does_file_or_directory_exist(catalog_info_file, storage_options=storage_options):
            raise FileNotFoundError(f"No catalog info found where expected: {str(catalog_info_file)}")



"""Utilities for find the pixels of higher orders that surround a given healpixel."""

import healpy as hp
import numpy as np

# see the documentation for get_edge() about this variable
_edge_vectors = [
    np.asarray([1, 3]),  # NE edge
    np.asarray([1]),  # E corner
    np.asarray([0, 1]),  # SE edge
    np.asarray([0]),  # S corner
    np.asarray([0, 2]),  # SW edge
    np.asarray([2]),  # W corner
    np.asarray([2, 3]),  # NW edge
    np.asarray([3]),  # N corner
]

# cache used by get_margin()
_suffixes = {}


def get_edge(d_order, pix, edge):
    """Get all the pixels at order kk+dk bordering pixel pix.
    See hipscat/pixel_math/README.md for more info.

    Args:
        dk (int): the change in k that we wish to find the margins for.
        pix (int): the healpix pixel to find margin pixels of.
        edge (int): the edge we want to find for the given pixel. (0-7)
            0: NE edge
            1: E corner
            2: SE edge
            3: S corner
            4: SW edge
            5: W corner
            6: NW edge
            7: N corner
    Returns:
        one-dimensional numpy array of long integers, filled with the healpix pixels
            at order kk+dk that border the given edge of pix.
    """
    if edge < 0 or edge > 7:
        raise ValueError("edge can only be values between 0 and 7 (see docstring)")

    if (edge, d_order) in _suffixes:
        pixel_edge = _suffixes[(edge, d_order)]
    else:
        # generate and cache the suffix:

        # generate all combinations of i,j,k,... suffixes for the requested edge
        # See https://stackoverflow.com/a/35608701
        grid = np.array(np.meshgrid(*[_edge_vectors[edge]] * d_order))
        pixel_edge = grid.T.reshape(-1, d_order)
        # bit-shift each suffix by the required number of bits
        pixel_edge <<= np.arange(2 * (d_order - 1), -2, -2)
        # sum them up row-wise, generating the suffix
        pixel_edge = pixel_edge.sum(axis=1)
        # cache for further reuse
        _suffixes[(edge, d_order)] = pixel_edge

    # append the 'pix' preffix
    pixel_edge = (pix << 2 * d_order) + pixel_edge
    return pixel_edge


def get_margin(order, pix, d_order):
    """Get all the pixels at order order+dk bordering pixel pix.
    See hipscat/pixel_math/README.md for more info.

    Args:
        order (int): the healpix order of pix.
        pix (int): the healpix pixel to find margin pixels of.
        d_order (int): the change in k that we wish to find the margins for.
            Must be greater than kk.
    Returns:
        one-dimensional numpy array of long integers, filled with the healpix pixels
            at order kk+dk that border pix.
    """
    if d_order < 1:
        raise ValueError("dk must be greater than order")

    nside = hp.order2nside(order)

    # get all neighboring pixels
    neighbors = hp.get_all_neighbours(nside, pix, nest=True)

    # get the healpix faces IDs of pix and the neighboring pixels
    _, _, pix_face = hp.pix2xyf(nside, pix, nest=True)
    _, _, faces = hp.pix2xyf(nside, neighbors, nest=True)

    # indices which tell get_edge() which edge/verted to return
    # for a given pixel. The default order is compatible with the
    # order returned by hp.get_all_neighbours().
    which = np.arange(8)

    # northern hemisphere; 90deg cw rotation for every +1 face increment
    if pix_face < 4:
        mask = faces < 4
        which[mask] += 2 * (faces - pix_face)[mask]
        which %= 8
    # southern hemisphere; 90deg ccw rotation for every +1 face increment
    elif pix_face >= 8:
        mask = faces >= 8
        which[mask] -= 2 * (faces - pix_face)[mask]
        which %= 8

    # get all edges/vertices
    # (making sure we skip -1 entries, for pixels with seven neighbors)
    margins = []
    for edge, pixel in zip(which, neighbors):
        if pixel != -1:
            margins.append(get_edge(d_order, pixel, edge))
    margins = np.concatenate(margins)
    return margins


def pixel_is_polar(order, pix):
    """Checks if a healpixel is a polar pixel.

    Args:
        order (int): the healpix order of the pixel to be checked.
        pix (int): the id of a healpixel to be checked. must be in the nested id scheme.
    Returns:
        tuple of a boolean and a string, the boolean indicating whether the pixel
            polar and the string denoting which pole it is ('North' or 'South')).
            string is empty in the case that the pixel isn't polar.
    """
    nside = hp.order2nside(order)
    npix = hp.nside2npix(nside)
    ring_pix = hp.nest2ring(nside, pix)

    # in the ring numbering scheme, the first and last 4 pixels are the poles.
    if ring_pix <= 3:
        return (True, "North")
    if ring_pix >= npix - 4:
        return (True, "South")
    return (False, "")



import os

import pandas as pd

from hipscat.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hipscat.io import file_io


def test_read_from_file(
    association_catalog_partition_join_file_cloud,
    example_cloud_storage_options,
    association_catalog_join_pixels,
):
    file_pointer = file_io.get_file_pointer_from_path(association_catalog_partition_join_file_cloud)
    info = PartitionJoinInfo.read_from_file(file_pointer, storage_options=example_cloud_storage_options)
    pd.testing.assert_frame_equal(info.data_frame, association_catalog_join_pixels)



import os

import pandas as pd
import pytest

from hipscat.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hipscat.io import file_io


def test_init(association_catalog_join_pixels):
    partition_join_info = PartitionJoinInfo(association_catalog_join_pixels)
    pd.testing.assert_frame_equal(partition_join_info.data_frame, association_catalog_join_pixels)


def test_wrong_columns(association_catalog_join_pixels):
    for column in PartitionJoinInfo.COLUMN_NAMES:
        join_pixels = association_catalog_join_pixels.copy()
        join_pixels = join_pixels.rename(columns={column: "wrong_name"})
        with pytest.raises(ValueError, match=column):
            PartitionJoinInfo(join_pixels)


def test_read_from_file(association_catalog_partition_join_file, association_catalog_join_pixels):
    file_pointer = file_io.get_file_pointer_from_path(association_catalog_partition_join_file)
    info = PartitionJoinInfo.read_from_file(file_pointer)
    pd.testing.assert_frame_equal(info.data_frame, association_catalog_join_pixels)


def test_read_from_missing_file(tmp_path):
    wrong_path = os.path.join(tmp_path, "wrong")
    file_pointer = file_io.get_file_pointer_from_path(wrong_path)
    with pytest.raises(FileNotFoundError):
        PartitionJoinInfo.read_from_file(file_pointer)



from typing import Any, Dict, List, Tuple, Union

import pandas as pd
from typing_extensions import TypeAlias

from hipscat.catalog.dataset import BaseCatalogInfo, Dataset
from hipscat.catalog.partition_info import PartitionInfo
from hipscat.io import FilePointer, file_io, paths
from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_tree import PixelTree
from hipscat.pixel_tree.pixel_tree_builder import PixelTreeBuilder

PixelInputTypes = Union[PartitionInfo, PixelTree, List[HealpixPixel]]


class HealpixDataset(Dataset):
    """A HiPSCat dataset partitioned with a HEALPix partitioning structure.

    Catalogs of this type are partitioned based on the ra and dec of the points with each partition
    containing points within a given HEALPix pixel. The files are in the form 'Norder=/Dir=/Npix=.parquet'.
    """

    CatalogInfoClass: TypeAlias = BaseCatalogInfo
    catalog_info: CatalogInfoClass

    def __init__(
        self,
        catalog_info: CatalogInfoClass,
        pixels: PixelInputTypes,
        catalog_path: str = None,
        storage_options: Union[Dict[Any, Any], None] = None,
    ) -> None:
        """Initializes a Catalog

        Args:
            catalog_info: CatalogInfo object with catalog metadata
            pixels: Specifies the pixels contained in the catalog. Can be either a
                list of HealpixPixel, `PartitionInfo object`, or a `PixelTree` object
            catalog_path: If the catalog is stored on disk, specify the location of the catalog
                Does not load the catalog from this path, only store as metadata
            storage_options: dictionary that contains abstract filesystem credentials
        """
        super().__init__(catalog_info, catalog_path, storage_options)
        self.partition_info = self._get_partition_info_from_pixels(pixels)
        self.pixel_tree = self._get_pixel_tree_from_pixels(pixels)

    def get_healpix_pixels(self) -> List[HealpixPixel]:
        """Get healpix pixel objects for all pixels contained in the catalog.

        Returns:
            List of HealpixPixel
        """
        return self.partition_info.get_healpix_pixels()

    @staticmethod
    def _get_partition_info_from_pixels(pixels: PixelInputTypes) -> PartitionInfo:
        if isinstance(pixels, PartitionInfo):
            return pixels
        if isinstance(pixels, PixelTree):
            return PartitionInfo.from_healpix(
                [
                    HealpixPixel(node.hp_order, node.hp_pixel)
                    for node in pixels.root_pixel.get_all_leaf_descendants()
                ]
            )
        if pd.api.types.is_list_like(pixels):
            return PartitionInfo.from_healpix(pixels)
        raise TypeError("Pixels must be of type PartitionInfo, PixelTree, or List[HealpixPixel]")

    @staticmethod
    def _get_pixel_tree_from_pixels(pixels: PixelInputTypes) -> PixelTree:
        if isinstance(pixels, PartitionInfo):
            return PixelTreeBuilder.from_healpix(pixels.get_healpix_pixels())
        if isinstance(pixels, PixelTree):
            return pixels
        if pd.api.types.is_list_like(pixels):
            return PixelTreeBuilder.from_healpix(pixels)
        raise TypeError("Pixels must be of type PartitionInfo, PixelTree, or List[HealpixPixel]")

    @classmethod
    def _read_args(
        cls, catalog_base_dir: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ) -> Tuple[CatalogInfoClass, PartitionInfo]:
        args = super()._read_args(catalog_base_dir, storage_options=storage_options)
        partition_info_file = paths.get_partition_info_pointer(catalog_base_dir)
        partition_info = PartitionInfo.read_from_file(partition_info_file, storage_options=storage_options)
        return args + (partition_info,)

    @classmethod
    def _check_files_exist(
        cls, catalog_base_dir: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ):
        super()._check_files_exist(catalog_base_dir, storage_options=storage_options)
        partition_info_file = paths.get_partition_info_pointer(catalog_base_dir)
        if not file_io.does_file_or_directory_exist(partition_info_file, storage_options=storage_options):
            raise FileNotFoundError(f"No partition info found where expected: {str(partition_info_file)}")



import pytest

from hipscat.catalog.association_catalog.association_catalog_info import AssociationCatalogInfo
from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo
from hipscat.catalog.dataset.catalog_info_factory import create_catalog_info, from_catalog_dir
from hipscat.catalog.index.index_catalog_info import IndexCatalogInfo
from hipscat.catalog.margin_cache.margin_cache_catalog_info import MarginCacheCatalogInfo
from hipscat.catalog.source_catalog.source_catalog_info import SourceCatalogInfo


def test_create_catalog_info_bad(base_catalog_info_data):
    base_catalog_info_data["catalog_type"] = "foo"
    with pytest.raises(ValueError, match="Unknown"):
        create_catalog_info(base_catalog_info_data)

    base_catalog_info_data.pop("catalog_type")
    with pytest.raises(ValueError, match="required"):
        create_catalog_info(base_catalog_info_data)


def test_create_catalog_info(base_catalog_info_data, catalog_info_data):
    catalog_info = create_catalog_info(base_catalog_info_data)
    assert catalog_info.catalog_name == "test_name"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, CatalogInfo)

    catalog_info = create_catalog_info(catalog_info_data)
    assert catalog_info.catalog_name == "test_name"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, CatalogInfo)


def test_create_catalog_info_association(association_catalog_info_data):
    catalog_info = create_catalog_info(association_catalog_info_data)
    assert catalog_info.catalog_name == "test_name"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, AssociationCatalogInfo)


def test_create_catalog_info_source(source_catalog_info, source_catalog_info_with_extra):
    catalog_info = create_catalog_info(source_catalog_info)
    assert catalog_info.catalog_name == "test_source"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, SourceCatalogInfo)

    catalog_info = create_catalog_info(source_catalog_info_with_extra)
    assert catalog_info.catalog_name == "test_source"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, SourceCatalogInfo)


def test_create_catalog_info_margin_cache(margin_cache_catalog_info):
    catalog_info = create_catalog_info(margin_cache_catalog_info)
    assert catalog_info.catalog_name == "test_margin"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, MarginCacheCatalogInfo)


def test_create_catalog_info_index(index_catalog_info, index_catalog_info_with_extra):
    catalog_info = create_catalog_info(index_catalog_info)
    assert catalog_info.catalog_name == "test_index"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, IndexCatalogInfo)

    catalog_info = create_catalog_info(index_catalog_info_with_extra)
    assert catalog_info.catalog_name == "test_index"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, IndexCatalogInfo)


def test_from_catalog_dir_object(small_sky_dir):
    catalog_info = from_catalog_dir(small_sky_dir)
    assert catalog_info.catalog_name == "small_sky"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, CatalogInfo)


def test_from_catalog_dir_source(source_catalog_info_file):
    catalog_info = from_catalog_dir(source_catalog_info_file)
    assert catalog_info.catalog_name == "small_sky_source_catalog"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, SourceCatalogInfo)


def test_from_catalog_dir_margin_cache(margin_cache_catalog_info_file):
    catalog_info = from_catalog_dir(margin_cache_catalog_info_file)
    assert catalog_info.catalog_name == "margin_cache"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, MarginCacheCatalogInfo)


def test_from_catalog_dir_index(index_catalog_info_file):
    catalog_info = from_catalog_dir(index_catalog_info_file)
    assert catalog_info.catalog_name == "index_catalog"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, IndexCatalogInfo)


def test_from_catalog_dir_association(association_catalog_path, association_catalog_info_file):
    catalog_info = from_catalog_dir(association_catalog_path)
    assert catalog_info.catalog_name == "small_sky_to_small_sky_order1"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, AssociationCatalogInfo)

    catalog_info = from_catalog_dir(association_catalog_info_file)
    assert catalog_info.catalog_name == "small_sky_to_small_sky_order1"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, AssociationCatalogInfo)



"""Set of convenience methods for testing file contents"""

import os
import re

import pyarrow as pa
import pytest

from hipscat.io.file_io.file_io import FilePointer, get_fs, load_text_file
from hipscat.io.file_io.file_pointer import does_file_or_directory_exist

# pylint: disable=missing-function-docstring, redefined-outer-name


@pytest.fixture
def assert_text_file_matches():
    def assert_text_file_matches(expected_lines, file_name, storage_options: dict = {}):
        """Convenience method to read a text file and compare the contents, line for line.

        When file contents get even a little bit big, it can be difficult to see
        the difference between an actual file and the expected contents without
        increased testing verbosity. This helper compares files line-by-line,
        using the provided strings or regular expressions.

        Notes:
            Because we check strings as regular expressions, you may need to escape some
            contents of `expected_lines`.

        Args:
            expected_lines(:obj:`string array`) list of strings, formatted as regular expressions.
            file_name (str): fully-specified path of the file to read
            storage_options (dict): dictionary of filesystem storage options
        """
        assert does_file_or_directory_exist(
            file_name, storage_options=storage_options
        ), f"file not found [{file_name}]"
        contents = load_text_file(file_name, storage_options=storage_options)

        assert len(expected_lines) == len(
            contents
        ), f"files not the same length ({len(contents)} vs {len(expected_lines)})"
        for i, expected in enumerate(expected_lines):
            assert re.match(expected, contents[i]), (
                f"files do not match at line {i+1} " f"(actual: [{contents[i]}] vs expected: [{expected}])"
            )

        # metadata_file.close()

    return assert_text_file_matches


@pytest.fixture
def copy_tree_fs_to_fs():
    def copy_tree_fs_to_fs(
        fs1_source: FilePointer,
        fs2_destination: FilePointer,
        storage_options1: dict = None,
        storage_options2: dict = None,
        existok=False,
        chunksize=1024 * 1024,
        verbose=False,
    ):
        """Recursive Copies directory from one filesystem to the other.

        Args:
            fs1_source: location of source directory to copy
            fs2_destination: location of destination directory to for fs1 to be written two
            storage_options1: dictionary that contains abstract filesystem1 credentials
            storage_options2: dictionary that contains abstract filesystem2 credentials
        """

        source_fs, source_fp = get_fs(fs1_source, storage_options=storage_options1)
        destination_fs, desintation_fp = get_fs(fs2_destination, storage_options=storage_options2)
        copy_dir(
            source_fs,
            source_fp,
            destination_fs,
            desintation_fp,
            chunksize=chunksize,
            existok=existok,
            verbose=verbose,
        )

    def copy_dir(
        source_fs,
        source_fp,
        destination_fs,
        desintation_fp,
        chunksize=1024 * 1024,
        existok=False,
        verbose=False,
    ):
        """Recursive method to copy directories and their contents.

        Args:
            fs1: fsspec.filesystem for the source directory contents
            fs1_pointer: source directory to copy content files
            fs2: fsspec.filesytem for destination directory
            fs2_pointer: destination directory for copied contents
        """
        destination_folder = os.path.join(desintation_fp, source_fp.split("/")[-1])
        if destination_folder[-1] != "/":
            destination_folder += "/"
        if not destination_fs.exists(destination_folder):
            destination_fs.makedirs(destination_folder, exist_ok=True)

        dir_contents = source_fs.listdir(source_fp)
        files = [x for x in source_fs.listdir(source_fp) if x["type"] == "file"]

        for _file in files:
            destination_fname = os.path.join(destination_folder, _file["name"].split("/")[-1])
            if not (destination_fs.exists(destination_fname) and existok):
                if verbose:
                    print(f"Creating destination folder: {destination_folder}")
                with source_fs.open(_file["name"], "rb") as source_file:
                    with destination_fs.open(destination_fname, "wb") as destination_file:
                        while True:
                            chunk = source_file.read(chunksize)
                            if not chunk:
                                break
                            destination_file.write(chunk)

        dirs = [x for x in dir_contents if x["type"] == "directory"]
        for _dir in dirs:
            copy_dir(source_fs, _dir["name"], destination_fs, destination_folder, chunksize=chunksize)

    return copy_tree_fs_to_fs


@pytest.fixture
def basic_catalog_parquet_metadata():
    return pa.schema(
        [
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.int64()),
            pa.field("dec_error", pa.int64()),
            pa.field("__index_level_0__", pa.int64()),
        ]
    )



import os

from hipscat.io.file_io import (
    append_paths_to_pointer,
    directory_has_contents,
    does_file_or_directory_exist,
    find_files_matching_path,
    get_directory_contents,
    get_file_pointer_from_path,
    is_regular_file,
)


def test_file_or_dir_exist(small_sky_dir_cloud, example_cloud_storage_options):
    small_sky_pointer = get_file_pointer_from_path(small_sky_dir_cloud)
    assert does_file_or_directory_exist(small_sky_pointer, storage_options=example_cloud_storage_options)
    catalog_info_string = os.path.join(small_sky_dir_cloud, "catalog_info.json")
    catalog_info_pointer = get_file_pointer_from_path(catalog_info_string)
    assert does_file_or_directory_exist(catalog_info_pointer, storage_options=example_cloud_storage_options)


def test_file_or_dir_exist_false(small_sky_dir_cloud, example_cloud_storage_options):
    small_sky_pointer = get_file_pointer_from_path(small_sky_dir_cloud + "incorrect file")
    assert not does_file_or_directory_exist(small_sky_pointer, storage_options=example_cloud_storage_options)


def test_is_regular_file(small_sky_dir_cloud, example_cloud_storage_options):
    partition_info_file = os.path.join(small_sky_dir_cloud, "partition_info.csv")
    assert is_regular_file(partition_info_file, storage_options=example_cloud_storage_options)

    assert not is_regular_file(small_sky_dir_cloud, storage_options=example_cloud_storage_options)

    partition_dir = os.path.join(small_sky_dir_cloud, "Norder=0")
    assert not is_regular_file(partition_dir, storage_options=example_cloud_storage_options)


def test_find_files_matching_path(small_sky_dir_cloud, example_cloud_storage_options):
    ## no_wildcard
    assert (
        len(
            find_files_matching_path(
                small_sky_dir_cloud, "partition_info.csv", storage_options=example_cloud_storage_options
            )
        )
        == 1
    )

    ## wilcard in the name
    assert (
        len(
            find_files_matching_path(
                small_sky_dir_cloud, "*.csv", storage_options=example_cloud_storage_options
            )
        )
        == 1
    )


def test_find_files_matching_path_directory(small_sky_order1_dir_cloud, example_cloud_storage_options):
    assert (
        len(
            find_files_matching_path(
                small_sky_order1_dir_cloud, storage_options=example_cloud_storage_options
            )
        )
        == 1
    )

    ## wildcard in directory - will match all files at indicated depth
    assert (
        len(
            find_files_matching_path(
                small_sky_order1_dir_cloud, "*", "*", "*", storage_options=example_cloud_storage_options
            )
        )
        == 4
    )


def test_directory_has_contents(small_sky_order1_dir_cloud, example_cloud_storage_options):
    assert directory_has_contents(small_sky_order1_dir_cloud, storage_options=example_cloud_storage_options)


def test_get_directory_contents(small_sky_order1_dir_cloud, example_cloud_storage_options):
    small_sky_contents = get_directory_contents(
        small_sky_order1_dir_cloud, include_protocol=True, storage_options=example_cloud_storage_options
    )
    assert len(small_sky_contents) == 4

    expected = [
        os.path.join(small_sky_order1_dir_cloud, "Norder=1"),
        os.path.join(small_sky_order1_dir_cloud, "catalog_info.json"),
        os.path.join(small_sky_order1_dir_cloud, "partition_info.csv"),
        os.path.join(small_sky_order1_dir_cloud, "point_map.fits"),
    ]

    assert small_sky_contents == expected



from hipscat.pixel_tree.pixel_alignment import align_trees
from hipscat.pixel_tree.pixel_tree import PixelTree


def assert_trees_equal(tree1: PixelTree, tree2: PixelTree):
    assert len(tree1.pixels) == len(tree2.pixels)
    for order, nodes_at_order in tree1.pixels.items():
        assert order in tree2.pixels
        assert len(tree2.pixels[order]) == len(nodes_at_order)
        for pixel, node in nodes_at_order.items():
            assert (order, pixel) in tree2
            tree2_node = tree2[(order, pixel)]
            assert node.node_type == tree2_node.node_type


def test_pixel_tree_alignment_same_tree(pixel_tree_1):
    alignment = align_trees(pixel_tree_1, pixel_tree_1, "inner")
    assert_trees_equal(pixel_tree_1, alignment.pixel_tree)


def test_pixel_tree_alignment_inner(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_inner):
    alignment = align_trees(pixel_tree_2, pixel_tree_3, "inner")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_inner)


def test_pixel_tree_alignment_left(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_left):
    alignment = align_trees(pixel_tree_2, pixel_tree_3, "left")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_left)


def test_pixel_tree_alignment_right(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_right):
    alignment = align_trees(pixel_tree_2, pixel_tree_3, "right")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_right)


def test_pixel_tree_alignment_outer(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_outer):
    alignment = align_trees(pixel_tree_2, pixel_tree_3, "outer")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_outer)



"""Tests of histogram calculations"""

import healpy as hp
import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

import hipscat.pixel_math as hist
from hipscat.pixel_math.healpix_pixel import HealpixPixel


def test_small_sky_same_pixel():
    """Test partitioning two objects into the same large bucket"""

    data = pd.DataFrame(
        data=[[700, 282.5, -58.5], [701, 299.5, -48.5]],
        columns=["id", "ra", "dec"],
    )

    result = hist.generate_histogram(
        data=data,
        highest_order=0,
    )

    assert len(result) == 12

    expected = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2]
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_column_names_error():
    """Test with non-default column names (without specifying column names)"""

    data = pd.DataFrame(
        data=[[700, 282.5, -58.5], [701, 299.5, -48.5]],
        columns=["id", "ra_mean", "dec_mean"],
    )

    with pytest.raises(ValueError, match="Invalid column names"):
        hist.generate_histogram(
            data=data,
            highest_order=0,
        )


def test_column_names():
    """Test with non-default column names"""
    data = pd.DataFrame(
        data=[[700, 282.5, -58.5], [701, 299.5, -48.5]],
        columns=["id", "ra_mean", "dec_mean"],
    )

    result = hist.generate_histogram(
        data=data,
        highest_order=0,
        ra_column="ra_mean",
        dec_column="dec_mean",
    )

    assert len(result) == 12

    expected = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2]
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_alignment_wrong_size():
    """Check that the method raises error when the input histogram is not the expected size."""
    initial_histogram = np.asarray([0, 0, 0, 0, 0, 0, 0, 0, 0, 131])
    with pytest.raises(ValueError, match="histogram is not the right size"):
        hist.generate_alignment(initial_histogram, highest_order=0, threshold=250)


def test_alignment_exceeds_threshold_order0():
    """Check that the method raises error when some pixel exceeds the threshold."""
    initial_histogram = np.asarray([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 131])
    with pytest.raises(ValueError, match="exceeds threshold"):
        hist.generate_alignment(initial_histogram, highest_order=0, threshold=20)


def test_alignment_lowest_order_too_large():
    """Check that the method raises error when some pixel exceeds the threshold."""
    initial_histogram = hist.empty_histogram(1)
    with pytest.raises(ValueError, match="lowest_order"):
        hist.generate_alignment(initial_histogram, highest_order=1, lowest_order=2, threshold=20)


def test_alignment_exceeds_threshold_order2():
    """Check that the method raises error when some pixel exceeds the threshold."""
    initial_histogram = hist.empty_histogram(2)
    filled_pixels = [4, 11, 14, 13, 5, 7, 8, 9, 11, 23, 4, 4, 17, 0, 1, 0]
    initial_histogram[176:] = filled_pixels[:]
    with pytest.raises(ValueError, match="exceeds threshold"):
        hist.generate_alignment(initial_histogram, highest_order=2, threshold=20)


def test_alignment_small_sky_order0():
    """Create alignment from small sky's distribution at order 0"""
    initial_histogram = np.asarray([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 131])
    result = hist.generate_alignment(initial_histogram, highest_order=0, threshold=250)

    expected = np.full(12, None)
    expected[11] = (0, 11, 131)

    npt.assert_array_equal(result, expected)


def test_alignment_small_sky_order1():
    """Create alignment from small sky's distribution at order 1"""
    initial_histogram = hist.empty_histogram(1)
    filled_pixels = [42, 29, 42, 18]
    initial_histogram[44:] = filled_pixels[:]
    result = hist.generate_alignment(initial_histogram, highest_order=1, threshold=250)

    expected = np.full(48, None)
    expected[44:] = [(0, 11, 131), (0, 11, 131), (0, 11, 131), (0, 11, 131)]

    npt.assert_array_equal(result, expected)


def test_alignment_small_sky_order2():
    """Create alignment from small sky's distribution at order 2"""
    initial_histogram = hist.empty_histogram(2)
    filled_pixels = [4, 11, 14, 13, 5, 7, 8, 9, 11, 23, 4, 4, 17, 0, 1, 0]
    initial_histogram[176:] = filled_pixels[:]
    result = hist.generate_alignment(initial_histogram, highest_order=2, threshold=250)

    expected = np.full(hp.order2npix(2), None)
    tuples = [
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
    ]
    expected[176:192] = tuples

    npt.assert_array_equal(result, expected)


@pytest.mark.timeout(5)
def test_alignment_even_sky():
    """Create alignment from an even distribution at order 7"""
    initial_histogram = np.full(hp.order2npix(7), 40)
    result = hist.generate_alignment(initial_histogram, highest_order=7, threshold=1_000)
    # everything maps to order 5, given the density
    for mapping in result:
        assert mapping[0] == 5

    result = hist.generate_alignment(initial_histogram, highest_order=7, lowest_order=7, threshold=1_000)
    # everything maps to order 7 (would be 5, but lowest of 7 is enforced)
    for mapping in result:
        assert mapping[0] == 7


def test_destination_pixel_map_order1():
    """Create destination pixel map for small sky at order 1"""

    alignment = np.full(48, None)
    alignment[44:] = [(0, 11, 131), (0, 11, 131), (0, 11, 131), (0, 11, 131)]

    initial_histogram = hist.empty_histogram(1)
    filled_pixels = [51, 29, 51, 0]
    initial_histogram[44:] = filled_pixels[:]

    expected = {tuple([0, 11, 131]): [44, 45, 46]}

    result = hist.generate_destination_pixel_map(initial_histogram, alignment)

    npt.assert_array_equal(result, expected)


def test_compute_pixel_map_order1():
    """Create destination pixel map for small sky at order 1"""

    initial_histogram = hist.empty_histogram(1)
    filled_pixels = [51, 29, 51, 0]
    initial_histogram[44:] = filled_pixels[:]

    expected = {HealpixPixel(0, 11): (131, [44, 45, 46])}

    result = hist.compute_pixel_map(initial_histogram, highest_order=1, threshold=150)

    npt.assert_array_equal(result, expected)

    expected = {
        HealpixPixel(1, 44): (51, [44]),
        HealpixPixel(1, 45): (29, [45]),
        HealpixPixel(1, 46): (51, [46]),
    }

    result = hist.compute_pixel_map(initial_histogram, highest_order=1, threshold=100)

    npt.assert_array_equal(result, expected)


@pytest.mark.timeout(5)
def test_compute_pixel_map_even_sky():
    """Create alignment from an even distribution at order 6"""
    initial_histogram = np.full(hp.order2npix(6), 200)
    result = hist.compute_pixel_map(initial_histogram, highest_order=6, threshold=1_000)
    # everything maps to order 5, given the density
    for mapping in result:
        assert mapping.order == 5


@pytest.mark.timeout(5)
def test_compute_pixel_map_even_sky_enforce_lowest():
    """Create pixel map for an even distribution, and enforce a lowest order bound."""
    initial_histogram = np.full(hp.order2npix(6), 10)
    result = hist.compute_pixel_map(initial_histogram, highest_order=6, lowest_order=4, threshold=1_000)
    # everything maps to order 4 (would be 0, but lowest of 4 is enforced)
    for mapping in result:
        assert mapping.order == 4


def test_compute_pixel_map_invalid_inputs():
    """Create destination pixel map for small sky at order 1"""

    initial_histogram = hist.empty_histogram(1)
    filled_pixels = [51, 29, 51, 0]
    initial_histogram[44:] = filled_pixels[:]

    ## Order doesn't match histogram length
    with pytest.raises(ValueError, match="histogram is not the right size"):
        hist.compute_pixel_map(initial_histogram, highest_order=2, threshold=150)

    ## Some bins exceed threshold
    with pytest.raises(ValueError, match="exceeds threshold"):
        hist.compute_pixel_map(initial_histogram, highest_order=1, threshold=30)

    ## lowest_order too large
    with pytest.raises(ValueError, match="lowest_order"):
        hist.compute_pixel_map(initial_histogram, highest_order=1, lowest_order=2, threshold=30)


def test_generate_constant_pixel_map():
    """Create constant pixel map for small sky data"""

    initial_histogram = np.asarray([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 131])
    expected = {HealpixPixel(0, 11): (131, [11])}

    result = hist.generate_constant_pixel_map(initial_histogram, constant_healpix_order=0)
    npt.assert_array_equal(result, expected)

    initial_histogram = hist.empty_histogram(1)
    filled_pixels = [51, 29, 51, 0]
    initial_histogram[44:] = filled_pixels[:]
    expected = {
        HealpixPixel(1, 44): (51, [44]),
        HealpixPixel(1, 45): (29, [45]),
        HealpixPixel(1, 46): (51, [46]),
    }

    result = hist.generate_constant_pixel_map(initial_histogram, constant_healpix_order=1)

    npt.assert_array_equal(result, expected)


def test_generate_constant_pixel_map_invalid_inputs():
    """Create destination pixel map for small sky at order 1"""

    initial_histogram = hist.empty_histogram(1)
    filled_pixels = [51, 29, 51, 0]
    initial_histogram[44:] = filled_pixels[:]

    ## Order doesn't match histogram length
    with pytest.raises(ValueError, match="histogram is not the right size"):
        hist.generate_constant_pixel_map(initial_histogram, constant_healpix_order=2)


def test_destination_map_matching_behavior():
    """Test that we get the same size destination pixel map, whether we compute
    directly, or from an existing alignment."""
    raw_histogram = hist.empty_histogram(2)
    raw_histogram[0:6] = 22946
    raw_histogram[64:79] = 185670

    alignment = hist.generate_alignment(
        raw_histogram,
        highest_order=2,
        threshold=1_000_000,
    )
    destination_pixel_map_a = hist.compute_pixel_map(
        raw_histogram,
        highest_order=2,
        threshold=1_000_000,
    )

    destination_pixel_map_b = hist.generate_destination_pixel_map(
        raw_histogram,
        alignment,
    )

    assert len(destination_pixel_map_a) == len(destination_pixel_map_b)



import pytest

from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_math.healpix_pixel_convertor import HealpixInputTypes, get_healpix_pixel


def return_healpix(pixel: HealpixInputTypes):
    return get_healpix_pixel(pixel)


def test_function_can_take_either_argument():
    order = 5
    pixel = 10
    for arg in [HealpixPixel(order, pixel), (order, pixel)]:
        returned_pixel = return_healpix(arg)
        assert isinstance(returned_pixel, HealpixPixel)
        assert returned_pixel.order == order
        assert returned_pixel.pixel == pixel


def test_function_raises_error_for_wrong_type():
    with pytest.raises(TypeError):
        return_healpix("test")


def test_function_raises_error_for_wrong_size_tuple():
    with pytest.raises(ValueError):
        return_healpix((1, 2, 3))



"""Container class to hold per-partition metadata"""
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd

from hipscat.io import FilePointer, file_io
from hipscat.pixel_math import HealpixPixel


class PartitionInfo:
    """Container class for per-partition info."""

    METADATA_ORDER_COLUMN_NAME = "Norder"
    METADATA_DIR_COLUMN_NAME = "Dir"
    METADATA_PIXEL_COLUMN_NAME = "Npix"

    def __init__(self, pixel_list: List[HealpixPixel]) -> None:
        self.pixel_list = pixel_list

    def get_healpix_pixels(self) -> List[HealpixPixel]:
        """Get healpix pixel objects for all pixels represented as partitions.

        Returns:
            List of HealpixPixel
        """
        return self.pixel_list

    def get_highest_order(self) -> int:
        """Get the highest healpix order for the dataset.

        Returns:
            int representing highest order.
        """
        max_pixel = np.max(self.pixel_list)
        return max_pixel.order

    def write_to_file(self, partition_info_file: FilePointer):
        """Write all partition data to CSV file.

        Args:
            partition_info_file: FilePointer to where the `partition_info.csv`
                file will be written
        """
        file_io.write_dataframe_to_csv(self.as_dataframe(), partition_info_file, index=False)

    @classmethod
    def read_from_file(
        cls, partition_info_file: FilePointer, storage_options: Union[Dict[Any, Any], None] = None
    ):
        """Read partition info from a `partition_info.csv` file to create an object

        Args:
            partition_info_file: FilePointer to the `partition_info.csv` file
            storage_options: dictionary that contains abstract filesystem credentials

        Returns:
            A `PartitionInfo` object with the data from the file
        """
        if not file_io.does_file_or_directory_exist(partition_info_file, storage_options=storage_options):
            raise FileNotFoundError(f"No partition info found where expected: {str(partition_info_file)}")

        data_frame = file_io.load_csv_to_pandas(partition_info_file, storage_options=storage_options)

        pixel_list = [
            HealpixPixel(order, pixel)
            for order, pixel in zip(
                data_frame[cls.METADATA_ORDER_COLUMN_NAME],
                data_frame[cls.METADATA_PIXEL_COLUMN_NAME],
            )
        ]

        return cls(pixel_list)

    def as_dataframe(self):
        """Construct a pandas dataframe for the partition info pixels.

        Returns:
            Dataframe with order, directory, and pixel info.
        """
        partition_info_dict = {
            PartitionInfo.METADATA_ORDER_COLUMN_NAME: [],
            PartitionInfo.METADATA_PIXEL_COLUMN_NAME: [],
            PartitionInfo.METADATA_DIR_COLUMN_NAME: [],
        }
        for pixel in self.pixel_list:
            partition_info_dict[PartitionInfo.METADATA_ORDER_COLUMN_NAME].append(pixel.order)
            partition_info_dict[PartitionInfo.METADATA_PIXEL_COLUMN_NAME].append(pixel.pixel)
            partition_info_dict[PartitionInfo.METADATA_DIR_COLUMN_NAME].append(
                int(pixel.pixel / 10_000) * 10_000
            )
        return pd.DataFrame.from_dict(partition_info_dict)

    @classmethod
    def from_healpix(cls, healpix_pixels: List[HealpixPixel]):
        """Create a partition info object from a list of constituent healpix pixels.

        Args:
            healpix_pixels: list of healpix pixels
        Returns:
            A `PartitionInfo` object with the same healpix pixels
        """
        return cls(healpix_pixels)



"""Two sample benchmarks to compute runtime and memory usage.

For more information on writing benchmarks:
https://asv.readthedocs.io/en/stable/writing_benchmarks.html."""

import healpy as hp
import numpy as np

import hipscat.pixel_math as hist
from hipscat.catalog import Catalog
from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_tree_builder import PixelTreeBuilder


def time_test_alignment_even_sky():
    """Create alignment from an even distribution at order 7"""
    initial_histogram = np.full(hp.order2npix(7), 40)
    result = hist.generate_alignment(initial_histogram, highest_order=7, threshold=1_000)
    # everything maps to order 5, given the density
    for mapping in result:
        assert mapping[0] == 5


def time_test_cone_filter_multiple_order():
    """Create a catalog cone filter where we have multiple orders in the catalog"""
    catalog_info = CatalogInfo(
        **{
            "catalog_name": "test_name",
            "catalog_type": "object",
            "total_rows": 10,
            "epoch": "J2000",
            "ra_column": "ra",
            "dec_column": "dec",
        }
    )
    pixels = [HealpixPixel(6, 30), HealpixPixel(7, 124), HealpixPixel(7, 5000)]
    catalog = Catalog(catalog_info, pixels)
    filtered_catalog = catalog.filter_by_cone(47.1, 6, 30)
    assert filtered_catalog.get_healpix_pixels() == [HealpixPixel(6, 30), HealpixPixel(7, 124)]


class Suite:
    def __init__(self) -> None:
        """Just initialize things"""
        self.pixel_list = None

    def setup(self):
        self.pixel_list = [HealpixPixel(8, pixel) for pixel in np.arange(100000)]

    def time_pixel_tree_creation(self):
        PixelTreeBuilder.from_healpix(self.pixel_list)



from typing import List

import healpy as hp
import numpy as np

from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree import PixelAlignment, PixelAlignmentType, align_trees
from hipscat.pixel_tree.pixel_tree import PixelTree
from hipscat.pixel_tree.pixel_tree_builder import PixelTreeBuilder


def filter_pixels_by_cone(pixel_tree: PixelTree, ra: float, dec: float, radius: float) -> List[HealpixPixel]:
    """Filter the leaf pixels in a pixel tree to return a partition_info dataframe with the pixels
    that overlap with a cone

    Args:
        ra (float): Right Ascension of the center of the cone in degrees
        dec (float): Declination of the center of the cone in degrees
        radius (float): Radius of the cone in degrees

    Returns:
        List of HealpixPixels representing only the pixels that overlap with the specified cone
    """
    max_order = max(pixel_tree.pixels.keys())
    cone_tree = _generate_cone_pixel_tree(ra, dec, radius, max_order)
    cone_alignment = align_trees(pixel_tree, cone_tree, alignment_type=PixelAlignmentType.INNER)
    pixels_df = cone_alignment.pixel_mapping[
        [PixelAlignment.PRIMARY_ORDER_COLUMN_NAME, PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME]
    ]
    filtered_pixels_df = pixels_df.drop_duplicates()

    pixel_list = [
        HealpixPixel(order, pixel)
        for order, pixel in zip(
            filtered_pixels_df[PixelAlignment.PRIMARY_ORDER_COLUMN_NAME],
            filtered_pixels_df[PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME],
        )
    ]
    return pixel_list


def _generate_cone_pixel_tree(ra: float, dec: float, radius: float, order: int):
    """Generates a pixel_tree filled with leaf nodes at a given order that overlap with a cone"""
    n_side = hp.order2nside(order)
    center_vec = hp.ang2vec(ra, dec, lonlat=True)
    radius_radians = np.radians(radius)
    cone_pixels = hp.query_disc(n_side, center_vec, radius_radians, inclusive=True, nest=True)
    pixel_list = [HealpixPixel(order, cone_pixel) for cone_pixel in cone_pixels]
    cone_tree = PixelTreeBuilder.from_healpix(pixel_list)
    return cone_tree



import json
import os

import numpy as np
import pandas as pd
import pytest

from hipscat.io.file_io import (
    delete_file,
    get_file_pointer_from_path,
    load_csv_to_pandas,
    load_json_file,
    load_parquet_to_pandas,
    make_directory,
    read_parquet_file_to_pandas,
    remove_directory,
    write_dataframe_to_csv,
    write_string_to_file,
)
from hipscat.io.file_io.file_pointer import does_file_or_directory_exist
from hipscat.io.paths import pixel_catalog_file


def test_make_directory(tmp_path):
    test_dir_path = os.path.join(tmp_path, "test_path")
    assert not does_file_or_directory_exist(test_dir_path)
    test_dir_pointer = get_file_pointer_from_path(test_dir_path)
    make_directory(test_dir_pointer)
    assert does_file_or_directory_exist(test_dir_path)


def test_make_existing_directory_raises(tmp_path):
    test_dir_path = os.path.join(tmp_path, "test_path")
    make_directory(test_dir_path)
    assert does_file_or_directory_exist(test_dir_path)
    test_dir_pointer = get_file_pointer_from_path(test_dir_path)
    with pytest.raises(OSError):
        make_directory(test_dir_pointer)


def test_make_existing_directory_existok(tmp_path):
    test_dir_path = os.path.join(tmp_path, "test_path")
    make_directory(test_dir_path)
    test_inner_dir_path = os.path.join(test_dir_path, "test_inner")
    make_directory(test_inner_dir_path)
    assert does_file_or_directory_exist(test_dir_path)
    assert does_file_or_directory_exist(test_inner_dir_path)
    test_dir_pointer = get_file_pointer_from_path(test_dir_path)
    make_directory(test_dir_pointer, exist_ok=True)
    assert does_file_or_directory_exist(test_dir_path)
    assert does_file_or_directory_exist(test_inner_dir_path)


def test_make_and_remove_directory(tmp_path):
    test_dir_path = os.path.join(tmp_path, "test_path")
    assert not does_file_or_directory_exist(test_dir_path)
    test_dir_pointer = get_file_pointer_from_path(test_dir_path)
    make_directory(test_dir_pointer)
    assert does_file_or_directory_exist(test_dir_path)
    remove_directory(test_dir_pointer)
    assert not does_file_or_directory_exist(test_dir_path)

    ## Directory no longer exists to be deleted.
    with pytest.raises(FileNotFoundError):
        remove_directory(test_dir_pointer)

    ## Directory doesn't exist, but shouldn't throw an error.
    remove_directory(test_dir_pointer, ignore_errors=True)
    assert not does_file_or_directory_exist(test_dir_path)


def test_write_string_to_file(tmp_path):
    test_file_path = os.path.join(tmp_path, "text_file.txt")
    test_file_pointer = get_file_pointer_from_path(test_file_path)
    test_string = "this is a test"
    write_string_to_file(test_file_pointer, test_string, encoding="utf-8")
    with open(test_file_path, "r", encoding="utf-8") as file:
        data = file.read()
        assert data == test_string
    delete_file(test_file_pointer)
    assert not does_file_or_directory_exist(test_file_pointer)


def test_load_json(small_sky_dir):
    catalog_info_path = os.path.join(small_sky_dir, "catalog_info.json")
    json_dict = None
    with open(catalog_info_path, "r", encoding="utf-8") as json_file:
        json_dict = json.load(json_file)
    catalog_info_pointer = get_file_pointer_from_path(catalog_info_path)
    loaded_json_dict = load_json_file(catalog_info_pointer, encoding="utf-8")
    assert loaded_json_dict == json_dict


def test_load_csv_to_pandas(small_sky_dir):
    partition_info_path = os.path.join(small_sky_dir, "partition_info.csv")
    csv_df = pd.read_csv(partition_info_path)
    partition_info_pointer = get_file_pointer_from_path(partition_info_path)
    loaded_df = load_csv_to_pandas(partition_info_pointer)
    pd.testing.assert_frame_equal(csv_df, loaded_df)


def test_load_parquet_to_pandas(small_sky_dir):
    pixel_data_path = pixel_catalog_file(small_sky_dir, 0, 11)
    parquet_df = pd.read_parquet(pixel_data_path)
    loaded_df = load_parquet_to_pandas(pixel_data_path)
    pd.testing.assert_frame_equal(parquet_df, loaded_df)


def test_write_df_to_csv(tmp_path):
    random_df = pd.DataFrame(np.random.randint(0, 100, size=(100, 4)), columns=list("ABCD"))
    test_file_path = os.path.join(tmp_path, "test.csv")
    test_file_pointer = get_file_pointer_from_path(test_file_path)
    write_dataframe_to_csv(random_df, test_file_pointer, index=False)
    loaded_df = pd.read_csv(test_file_path)
    pd.testing.assert_frame_equal(loaded_df, random_df)


def test_read_parquet_data(tmp_path):
    random_df = pd.DataFrame(np.random.randint(0, 100, size=(100, 4)), columns=list("ABCD"))
    test_file_path = os.path.join(tmp_path, "test.parquet")
    random_df.to_parquet(test_file_path)
    file_pointer = get_file_pointer_from_path(test_file_path)
    dataframe = read_parquet_file_to_pandas(file_pointer)
    pd.testing.assert_frame_equal(dataframe, random_df)



import pytest

from hipscat.catalog import Catalog
from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_node import PixelNode
from hipscat.pixel_tree.pixel_node_type import PixelNodeType

# pylint: disable= redefined-outer-name


@pytest.fixture
def root_pixel_node_data():
    return {
        "pixel": HealpixPixel(-1, -1),
        "node_type": PixelNodeType.ROOT,
        "parent": None,
    }


@pytest.fixture
def root_pixel_node(root_pixel_node_data):
    return PixelNode(**root_pixel_node_data)


@pytest.fixture
def inner_pixel_node_data(root_pixel_node):
    return {
        "pixel": HealpixPixel(0, 1),
        "node_type": PixelNodeType.INNER,
        "parent": root_pixel_node,
    }


@pytest.fixture
def inner_pixel_node(inner_pixel_node_data):
    return PixelNode(**inner_pixel_node_data)


@pytest.fixture
def leaf_pixel_node_data(inner_pixel_node):
    return {
        "pixel": HealpixPixel(1, 12),
        "node_type": PixelNodeType.LEAF,
        "parent": inner_pixel_node,
    }


@pytest.fixture
def leaf_pixel_node(leaf_pixel_node_data):
    return PixelNode(**leaf_pixel_node_data)


@pytest.fixture
def small_sky_catalog(small_sky_dir):
    return Catalog.read_from_hipscat(small_sky_dir)


@pytest.fixture()
def small_sky_pixels():
    return [
        HealpixPixel(0, 11),
    ]


@pytest.fixture
def small_sky_order1_catalog(small_sky_order1_dir):
    return Catalog.read_from_hipscat(small_sky_order1_dir)


@pytest.fixture()
def small_sky_order1_pixels():
    return [
        HealpixPixel(1, 44),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
        HealpixPixel(1, 47),
    ]



import dataclasses
import json

from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.io import file_io


def test_catalog_info(catalog_info_data, assert_catalog_info_matches_dict):
    info = CatalogInfo(**catalog_info_data)
    assert_catalog_info_matches_dict(info, catalog_info_data)


def test_catalog_info_defaults(base_catalog_info_data, assert_catalog_info_matches_dict):
    info = CatalogInfo(**base_catalog_info_data)
    actual_catalog_info = base_catalog_info_data.copy()
    actual_catalog_info["epoch"] = "J2000"
    actual_catalog_info["ra_column"] = "ra"
    actual_catalog_info["dec_column"] = "dec"
    assert_catalog_info_matches_dict(info, actual_catalog_info)


def test_str(catalog_info_data):
    correct_string = ""
    for name, value in catalog_info_data.items():
        correct_string += f"  {name} {value}\n"
    cat_info = CatalogInfo(**catalog_info_data)
    assert str(cat_info) == correct_string


def test_read_from_file(catalog_info_file, assert_catalog_info_matches_dict):
    cat_info_fp = file_io.get_file_pointer_from_path(catalog_info_file)
    catalog_info = CatalogInfo.read_from_metadata_file(cat_info_fp)
    for column in [
        "catalog_name",
        "catalog_type",
        "total_rows",
        "epoch",
        "ra_column",
        "dec_column",
    ]:
        assert column in dataclasses.asdict(catalog_info)

    with open(catalog_info_file, "r", encoding="utf-8") as cat_info_file:
        catalog_info_json = json.load(cat_info_file)
        assert_catalog_info_matches_dict(catalog_info, catalog_info_json)



"""Tests of catalog functionality"""

import os

import pytest

from hipscat.catalog import Catalog, PartitionInfo
from hipscat.io.file_io import file_io


def test_load_catalog_small_sky(small_sky_dir_cloud, example_cloud_storage_options):
    """Instantiate a catalog with 1 pixel"""
    cat = Catalog.read_from_hipscat(small_sky_dir_cloud, storage_options=example_cloud_storage_options)

    assert cat.catalog_name == "small_sky"
    assert len(cat.get_healpix_pixels()) == 1


def test_load_catalog_small_sky_order1(small_sky_order1_dir_cloud, example_cloud_storage_options):
    """Instantiate a catalog with 4 pixels"""
    cat = Catalog.read_from_hipscat(small_sky_order1_dir_cloud, storage_options=example_cloud_storage_options)

    assert cat.catalog_name == "small_sky_order1"
    assert len(cat.get_healpix_pixels()) == 4


def test_cone_filter(small_sky_order1_catalog_cloud):
    filtered_catalog = small_sky_order1_catalog_cloud.filter_by_cone(315, -66.443, 0.1)
    assert len(filtered_catalog.partition_info.data_frame) == 1
    assert filtered_catalog.partition_info.data_frame[PartitionInfo.METADATA_PIXEL_COLUMN_NAME][0] == 44
    assert filtered_catalog.partition_info.data_frame[PartitionInfo.METADATA_ORDER_COLUMN_NAME][0] == 1
    assert (1, 44) in filtered_catalog.pixel_tree
    assert len(filtered_catalog.pixel_tree.pixels[1]) == 1
    assert filtered_catalog.catalog_info.total_rows is None


def test_cone_filter_big(small_sky_order1_catalog_cloud):
    filtered_catalog = small_sky_order1_catalog_cloud.filter_by_cone(315, -66.443, 30)
    assert len(filtered_catalog.partition_info.data_frame) == 4
    assert (1, 44) in filtered_catalog.pixel_tree
    assert (1, 45) in filtered_catalog.pixel_tree
    assert (1, 46) in filtered_catalog.pixel_tree
    assert (1, 47) in filtered_catalog.pixel_tree


def test_cone_filter_empty(small_sky_order1_catalog_cloud):
    filtered_catalog = small_sky_order1_catalog_cloud.filter_by_cone(0, 0, 0.1)
    assert len(filtered_catalog.partition_info.data_frame) == 0
    assert len(filtered_catalog.pixel_tree) == 1


def test_empty_directory(tmp_dir_cloud, example_cloud_storage_options):
    """Test loading empty or incomplete data"""
    catalog_path = os.path.join(tmp_dir_cloud, "empty")

    ## Path exists but there's nothing there
    with pytest.raises(FileNotFoundError):
        Catalog.read_from_hipscat(catalog_path, storage_options=example_cloud_storage_options)

    ## catalog_info file exists - getting closer
    file_name = os.path.join(catalog_path, "catalog_info.json")
    file_io.write_string_to_file(
        file_name,
        string='{"catalog_name":"empty", "catalog_type":"source"}',
        storage_options=example_cloud_storage_options,
    )

    with pytest.raises(FileNotFoundError):
        Catalog.read_from_hipscat(catalog_path, storage_options=example_cloud_storage_options)

    ## partition_info file exists - enough to create a catalog
    file_name = os.path.join(catalog_path, "partition_info.csv")
    file_io.write_string_to_file(file_name, string="foo", storage_options=example_cloud_storage_options)

    catalog = Catalog.read_from_hipscat(catalog_path, storage_options=example_cloud_storage_options)
    assert catalog.catalog_name == "empty"

    file_io.delete_file(
        os.path.join(catalog_path, "catalog_info.json"), storage_options=example_cloud_storage_options
    )
    file_io.delete_file(
        os.path.join(catalog_path, "partition_info.csv"), storage_options=example_cloud_storage_options
    )



# pylint: disable=duplicate-code

from __future__ import annotations

from typing import List

from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_math.healpix_pixel_convertor import HealpixInputTypes, get_healpix_pixel
from hipscat.pixel_tree.pixel_node import PixelNode
from hipscat.pixel_tree.pixel_node_type import PixelNodeType
from hipscat.pixel_tree.pixel_tree import PixelTree


class PixelTreeBuilder:
    """Build a PixelTree

    Initially a root node is created when the builder is initialized.
    Nodes can then be added to the tree.
    To create a pixel tree object once the tree is built, call the `build` method

    """

    def __init__(self):
        self.root_pixel = PixelNode((-1, -1), PixelNodeType.ROOT, None)
        self.pixels = {-1: {-1: self.root_pixel}}

    def build(self) -> PixelTree:
        """Build a `PixelTree` object from the nodes created

        Returns:
            The pixel tree with the nodes created in the builder
        """
        return PixelTree(self.root_pixel, self.pixels)

    @staticmethod
    def from_healpix(healpix_pixels: List[HealpixPixel]) -> PixelTree:
        """Build a tree from a list of constituent healpix pixels

        Args:
            healpix_pixels: list of healpix pixels

        Returns:
            The pixel tree with the leaf pixels specified in the list
        """
        builder = PixelTreeBuilder()
        for pixel in healpix_pixels:
            builder.create_node_and_parent_if_not_exist((pixel.order, pixel.pixel), PixelNodeType.LEAF)
        return builder.build()

    def contains(self, pixel: HealpixInputTypes) -> bool:
        """Check if tree contains a node at a given order and pixel

        Args:
            pixel: HEALPix pixel to check. Either of type `HealpixPixel`
                or a tuple of (order, pixel)

        Returns:
            True if the tree contains the pixel, False if not
        """
        if isinstance(pixel, HealpixPixel):
            return pixel.order in self.pixels and pixel.pixel in self.pixels[pixel.order]
        if isinstance(pixel, tuple):
            return pixel[0] in self.pixels and pixel[1] in self.pixels[pixel[0]]
        raise TypeError("pixel must be either a HealpixPixel object or a Tuple of (order, pixel)")

    def __contains__(self, item):
        return self.contains(item)

    def get_node(self, pixel: HealpixInputTypes) -> PixelNode | None:
        """Get the node at a given pixel

        Args:
            pixel: HEALPix pixel to get. Either of type `HealpixPixel`
                or a tuple of (order, pixel)

        Returns:
            The PixelNode at the index, or None if a node does not exist
        """
        pixel = get_healpix_pixel(pixel)
        if self.contains(pixel):
            return self.pixels[pixel.order][pixel.pixel]
        return None

    def __getitem__(self, item) -> PixelNode:
        return self.get_node(item)

    def create_node_and_parent_if_not_exist(self, pixel: HealpixInputTypes, node_type: PixelNodeType):
        """Creates a node and adds to `self.pixels` in the tree, and recursively creates parent
        node if parent does not exist

        Args:
            pixel: HEALPix pixel to get. Either of type `HealpixPixel`
                or a tuple of (order, pixel)
            node_type: Node type of the node to create
        """
        pixel = get_healpix_pixel(pixel)
        if self.contains(pixel):
            raise ValueError("Incorrectly configured catalog: catalog contains duplicate pixels")

        if pixel.order == 0:
            self.create_node(pixel, node_type, self.root_pixel)
            return

        parents_to_add = []
        for delta_order in range(1, pixel.order + 1):
            parent_order = pixel.order - delta_order
            parent_pixel = pixel.pixel >> (2 * delta_order)
            if not self.contains((parent_order, parent_pixel)):
                parents_to_add.insert(0, (parent_order, parent_pixel))
            else:
                break

        for add_order, add_pixel in parents_to_add:
            parent_order = add_order - 1
            parent_pixel = add_pixel >> 2 if add_order > 0 else -1
            self.create_node(
                (add_order, add_pixel), PixelNodeType.INNER, self.pixels[parent_order][parent_pixel]
            )
        parent_order = pixel.order - 1
        parent_pixel = pixel.pixel >> 2 if pixel.order > 0 else -1

        parent = self.pixels[parent_order][parent_pixel]

        if parent.node_type != PixelNodeType.INNER:
            raise ValueError(
                "Incorrectly configured catalog: catalog contains pixels defined at multiple orders"
            )

        self.create_node(pixel, node_type, parent)

    def create_node(
        self,
        pixel: HealpixInputTypes,
        node_type: PixelNodeType,
        parent: PixelNode = None,
        replace_existing_node=False,
    ):
        """Create a node and add to `self.pixels` in the tree

        Args:
            pixel: HEALPix pixel to get. Either of type `HealpixPixel`
                or a tuple of (order, pixel)
            node_type: Node type of the node to create
            parent: (Default: None) Parent node of the node to create. If None, will get parent from
                existing node in tree at correct position based on pixel
            replace_existing_node: (Default: False) If True and node at `pixel` already exists,
                the existing node will be removed and replaced with the new node.
        """
        pixel = get_healpix_pixel(pixel)
        if parent is None:
            if pixel.order == 0:
                parent = self.root_pixel
            else:
                parent_pixel = pixel.convert_to_lower_order(delta_order=1)
                if parent_pixel not in self:
                    raise ValueError(
                        f"Cannot create node at {str(pixel)}, "
                        f"no parent node exists at {str(parent_pixel)}"
                    )
                parent = self[parent_pixel]
        if parent.node_type == PixelNodeType.LEAF:
            raise ValueError(
                f"Cannot create node at {str(pixel)}, "
                f"parent node at {str(parent_pixel)} has node type leaf"
            )
        node_to_replace = None
        if pixel in self:
            if not replace_existing_node:
                raise ValueError(f"Cannot create node at {str(pixel)}, node already exists")
            node_to_replace = self[pixel]
            parent.remove_child_link(node_to_replace)
        node = PixelNode(pixel, node_type, parent)
        if pixel.order not in self.pixels:
            self.pixels[pixel.order] = {}
        self.pixels[pixel.order][pixel.pixel] = node
        if node_to_replace is not None:
            if node.node_type != PixelNodeType.LEAF:
                for child in node_to_replace.children:
                    child.parent = node
                    node.add_child_node(child)
            else:
                for child in node_to_replace.children:
                    self._remove_node_and_children_from_tree(child.pixel)

    def remove_node(self, pixel: HealpixInputTypes):
        """Remove node in tree

        Removes a node at a specified pixel in the tree. Raises a `ValueError` if no node at
        specified pixel.

        Args:
            pixel: HEALPix pixel to remove the node at

        Raises:
            ValueError: No node in tree at pixel
        """
        pixel = get_healpix_pixel(pixel)
        if pixel not in self:
            raise ValueError(f"No node at pixel {str(pixel)}")
        node = self[pixel]
        node.parent.remove_child_link(node)
        self._remove_node_and_children_from_tree(pixel)

    def _remove_node_and_children_from_tree(self, pixel: HealpixPixel):
        node = self.pixels[pixel.order].pop(pixel.pixel)
        if len(self.pixels[pixel.order]) == 0:
            self.pixels.pop(pixel.order)
        for child_node in node.children:
            self._remove_node_and_children_from_tree(child_node.pixel)

    def add_all_descendants_from_node(self, node: PixelNode):
        """Adds all descendents from a given node to the current tree

        Used to make the current tree being built mirror the subtree from the specified node in
        another tree.

        A node at the specified position must exist in the current tree with the correct node type.

        Args:
            node: The node from another tree to match

        Raises:
            ValueError: No node exists in the current tree at the same location as the specified
            node
        """
        if node.pixel not in self:
            raise ValueError("No node in tree matching given node")
        nodes_to_add = list(node.children)
        while len(nodes_to_add) > 0:
            node = nodes_to_add.pop(0)
            nodes_to_add += node.children
            self.create_node(node.pixel, node.node_type)

    def split_leaf_to_match_partitioning(self, node_to_match: PixelNode):
        """Split a given leaf node into higher order nodes to match another node's partitioning

        Args:
            node_to_match: The node in another tree to match the partitioning of

        Raises:
            ValueError if no node in tree at the same position as node_to_match,
             or node in tree isn't a leaf node.
        """
        if node_to_match.pixel not in self:
            raise ValueError("No node in tree matching given node")
        if self[node_to_match.pixel].node_type != PixelNodeType.LEAF:
            raise ValueError("Node in tree is not a leaf node")
        nodes_to_add = list(node_to_match.children)
        while len(nodes_to_add) > 0:
            node = nodes_to_add.pop(0)
            nodes_to_add += node.children
            parent_node = self[node.parent.pixel]
            if parent_node.node_type == PixelNodeType.LEAF:
                parent_node.node_type = PixelNodeType.INNER
                for child_pixel in parent_node.pixel.convert_to_higher_order(delta_order=1):
                    self.create_node(child_pixel, PixelNodeType.LEAF, parent=parent_node)



import pytest

from hipscat.pixel_math.healpix_pixel import HealpixPixel
from hipscat.pixel_math.hipscat_id import HIPSCAT_ID_HEALPIX_ORDER


def test_pixels_equal():
    order = 3
    pixel = 42
    pix1 = HealpixPixel(order=order, pixel=pixel)
    pix2 = HealpixPixel(order=order, pixel=pixel)
    assert pix1 == pix2
    assert pix1.order == pix2.order
    assert pix1.pixel == pix2.pixel


def test_order_greater_than_max_order_fails():
    with pytest.raises(ValueError):
        HealpixPixel(order=20, pixel=0)


def test_equal_pixel_hash_equal():
    order = 3
    pixel = 42
    test_string = "testing"
    pix1 = HealpixPixel(order=order, pixel=pixel)
    test_dict = {}
    test_dict[pix1] = test_string
    pix2 = HealpixPixel(order=order, pixel=pixel)
    assert pix1 == pix2
    assert pix2 in test_dict
    assert test_dict[pix2] == test_string


def test_pixel_str_and_repr():
    order = 3
    pixel = 42
    test_string = f"Order: {order}, Pixel: {pixel}"
    pix = HealpixPixel(order=order, pixel=pixel)
    assert str(pix) == test_string
    assert repr(pix) == test_string


@pytest.mark.parametrize(
    "order, pixel, final_order, final_pixel",
    [
        (3, 10, 2, 2),
        (6, 1033, 5, 258),
        (4, 0, 3, 0),
        (5, 400, 2, 6),
        (2, 4, 0, 0),
        (3, 10, 3, 10),
    ],
)
def test_convert_lower_order(order, pixel, final_order, final_pixel):
    delta = order - final_order
    pixel = HealpixPixel(order, pixel)
    final_pix = pixel.convert_to_lower_order(delta)
    assert final_pix.order == final_order
    assert final_pix.pixel == final_pixel


def test_convert_lower_order_fails_below_zero():
    order = 4
    pixel = 3
    pixel = HealpixPixel(order, pixel)
    with pytest.raises(ValueError):
        pixel.convert_to_lower_order(order + 1)


def test_convert_lower_order_fails_negative():
    order = 4
    pixel = 3
    pixel = HealpixPixel(order, pixel)
    with pytest.raises(ValueError):
        pixel.convert_to_lower_order(-1)


@pytest.mark.parametrize(
    "order, pixel, delta_order", [(3, 10, 1), (6, 1033, 3), (4, 0, 1), (5, 400, 2), (2, 4, 0), (0, 10, 2)]
)
def test_convert_higher_order(order, pixel, delta_order):
    converted_pixels = HealpixPixel(order, pixel).convert_to_higher_order(delta_order)
    final_order = order + delta_order
    for final_pixel in range(pixel * 4**delta_order, (pixel + 1) * 4**delta_order):
        assert HealpixPixel(final_order, final_pixel) in converted_pixels


def test_convert_higher_order_fails_above_limit():
    order = 4
    pixel = 3
    pixel = HealpixPixel(order, pixel)
    with pytest.raises(ValueError):
        pixel.convert_to_higher_order(HIPSCAT_ID_HEALPIX_ORDER - order + 1)


def test_convert_higher_order_fails_negative():
    order = 4
    pixel = 3
    pixel = HealpixPixel(order, pixel)
    with pytest.raises(ValueError):
        pixel.convert_to_higher_order(-1)



"""Methods for creating partitioned data paths"""
from __future__ import annotations

from hipscat.io.file_io.file_pointer import FilePointer, append_paths_to_pointer

ORDER_DIRECTORY_PREFIX = "Norder"
DIR_DIRECTORY_PREFIX = "Dir"
PIXEL_DIRECTORY_PREFIX = "Npix"
JOIN_ORDER_DIRECTORY_PREFIX = "join_Norder"
JOIN_DIR_DIRECTORY_PREFIX = "join_Dir"
JOIN_PIXEL_DIRECTORY_PREFIX = "join_Npix"

CATALOG_INFO_FILENAME = "catalog_info.json"
PARTITION_INFO_FILENAME = "partition_info.csv"
PARTITION_JOIN_INFO_FILENAME = "partition_join_info.csv"
PROVENANCE_INFO_FILENAME = "provenance_info.json"
PARQUET_METADATA_FILENAME = "_metadata"
PARQUET_COMMON_METADATA_FILENAME = "_common_metadata"
POINT_MAP_FILENAME = "point_map.fits"


def pixel_directory(
    catalog_base_dir: FilePointer,
    pixel_order: int,
    pixel_number: int | None = None,
    directory_number: int | None = None,
) -> FilePointer:
    """Create path pointer for a pixel directory. This will not create the directory.

    One of pixel_number or directory_number is required. The directory name will
    take the HiPS standard form of:

        <catalog_base_dir>/Norder=<pixel_order>/Dir=<directory number>

    Where the directory number is calculated using integer division as:

        (pixel_number/10000)*10000

    Args:
        catalog_base_dir (FilePointer): base directory of the catalog (includes catalog name)
        pixel_order (int): the healpix order of the pixel
        directory_number (int): directory number
        pixel_number (int): the healpix pixel
    Returns:
        FilePointer directory name
    """
    norder = int(pixel_order)
    if directory_number is not None:
        ndir = directory_number
    elif pixel_number is not None:
        npix = int(pixel_number)
        ndir = int(npix / 10_000) * 10_000
    else:
        raise ValueError("One of pixel_number or directory_number is required to create pixel directory")
    return create_hive_directory_name(
        catalog_base_dir,
        [ORDER_DIRECTORY_PREFIX, DIR_DIRECTORY_PREFIX],
        [norder, ndir],
    )


def pixel_catalog_file(catalog_base_dir: FilePointer, pixel_order: int, pixel_number: int) -> FilePointer:
    """Create path *pointer* for a pixel catalog file. This will not create the directory or file.

    The catalog file name will take the HiPS standard form of:

        <catalog_base_dir>/Norder=<pixel_order>/Dir=<directory number>/Npix=<pixel_number>.parquet

    Where the directory number is calculated using integer division as:

        (pixel_number/10000)*10000

    Args:
        catalog_base_dir (FilePointer): base directory of the catalog (includes catalog name)
        pixel_order (int): the healpix order of the pixel
        pixel_number (int): the healpix pixel
    Returns:
        string catalog file name
    """
    return create_hive_parquet_file_name(
        catalog_base_dir,
        [ORDER_DIRECTORY_PREFIX, DIR_DIRECTORY_PREFIX, PIXEL_DIRECTORY_PREFIX],
        [int(pixel_order), int(pixel_number / 10_000) * 10_000, int(pixel_number)],
    )


def create_hive_directory_name(base_dir, partition_token_names, partition_token_values):
    """Create path *pointer* for a directory with hive partitioning naming.
    This will not create the directory.

    The directory name will have the form of:

        <catalog_base_dir>/<name_1>=<value_1>/.../<name_n>=<value_n>

    Args:
        catalog_base_dir (FilePointer): base directory of the catalog (includes catalog name)
        partition_token_names (list[string]): list of partition name parts.
        partition_token_values (list[string]): list of partition values that
            correspond to the token name parts.
    """
    partition_tokens = [
        f"{name}={value}" for name, value in zip(partition_token_names, partition_token_values)
    ]
    return append_paths_to_pointer(base_dir, *partition_tokens)


def create_hive_parquet_file_name(base_dir, partition_token_names, partition_token_values):
    """Create path *pointer* for a single parquet with hive partitioning naming.

    The file name will have the form of:

        <catalog_base_dir>/<name_1>=<value_1>/.../<name_n>=<value_n>.parquet

    Args:
        catalog_base_dir (FilePointer): base directory of the catalog (includes catalog name)
        partition_token_names (list[string]): list of partition name parts.
        partition_token_values (list[string]): list of partition values that
            correspond to the token name parts.
    """
    partition_tokens = [
        f"{name}={value}" for name, value in zip(partition_token_names, partition_token_values)
    ]
    return f"{append_paths_to_pointer(base_dir, *partition_tokens)}.parquet"


def get_catalog_info_pointer(catalog_base_dir: FilePointer) -> FilePointer:
    """Get file pointer to `catalog_info.json` metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `catalog_info.json` file
    """
    return append_paths_to_pointer(catalog_base_dir, CATALOG_INFO_FILENAME)


def get_partition_info_pointer(catalog_base_dir: FilePointer) -> FilePointer:
    """Get file pointer to `partition_info.csv` metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `partition_info.csv` file
    """
    return append_paths_to_pointer(catalog_base_dir, PARTITION_INFO_FILENAME)


def get_provenance_pointer(catalog_base_dir: FilePointer) -> FilePointer:
    """Get file pointer to `provenance_info.json` metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `provenance_info.json` file
    """
    return append_paths_to_pointer(catalog_base_dir, PROVENANCE_INFO_FILENAME)


def get_common_metadata_pointer(catalog_base_dir: FilePointer) -> FilePointer:
    """Get file pointer to `_metadata` parquet metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `_metadata` file
    """
    return append_paths_to_pointer(catalog_base_dir, PARQUET_COMMON_METADATA_FILENAME)


def get_parquet_metadata_pointer(catalog_base_dir: FilePointer) -> FilePointer:
    """Get file pointer to `_common_metadata` parquet metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `_common_metadata` file
    """
    return append_paths_to_pointer(catalog_base_dir, PARQUET_METADATA_FILENAME)


def get_point_map_file_pointer(catalog_base_dir: FilePointer) -> FilePointer:
    """Get file pointer to `point_map.fits` FITS image file.

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `point_map.fits` FITS image file.
    """
    return append_paths_to_pointer(catalog_base_dir, POINT_MAP_FILENAME)


def get_partition_join_info_pointer(catalog_base_dir: FilePointer) -> FilePointer:
    """Get file pointer to `partition_join_info.csv` association metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `partition_join_info.csv` association metadata file
    """
    return append_paths_to_pointer(catalog_base_dir, PARTITION_JOIN_INFO_FILENAME)



from __future__ import annotations

import dataclasses
import os
from dataclasses import dataclass, field
from typing import Any, Dict, List, Union

import yaml
from typing_extensions import Self

from hipscat.catalog.dataset import catalog_info_factory
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo
from hipscat.io import file_io


@dataclass
class AlmanacInfo:
    """Container for parsed almanac information."""

    file_path: str = ""
    storage_options: Union[Dict[Any, Any], None] = field(default_factory=dict)
    namespace: str = ""
    catalog_path: str = ""
    catalog_name: str = ""
    catalog_type: str = ""
    primary: str | None = None
    join: str | None = None
    primary_link: Self | None = None
    join_link: Self | None = None
    sources: List[Self] = field(default_factory=list)
    objects: List[Self] = field(default_factory=list)
    margins: List[Self] = field(default_factory=list)
    associations: List[Self] = field(default_factory=list)
    associations_right: List[Self] = field(default_factory=list)
    indexes: List[Self] = field(default_factory=list)

    creators: List[str] = field(default_factory=list)
    description: str = ""
    version: str = ""
    deprecated: str = ""

    catalog_info: dict = field(default_factory=dict)

    catalog_info_object: BaseCatalogInfo | None = None

    def __post_init__(self):
        if len(self.catalog_info):
            self.catalog_info_object = catalog_info_factory.create_catalog_info(self.catalog_info)
            if self.catalog_info and "primary_catalog" in self.catalog_info and not self.primary:
                self.primary = self.catalog_info["primary_catalog"]
            if self.catalog_info and "join_catalog" in self.catalog_info and not self.join:
                self.join = self.catalog_info["join_catalog"]

        ## Allows use of $HIPSCAT_DEFAULT_DIR in paths
        self.catalog_path = os.path.expandvars(self.catalog_path)

    @staticmethod
    def get_default_dir() -> str:
        """Fetch the default directory for environment variables.

        This is set via the environment variable: HIPSCAT_ALMANAC_DIR

        To set this in a linux-like environment, use a command like:

            export HIPSCAT_ALMANAC_DIR=/data/path/to/almanacs

        This will also attempt to expand any environment variables WITHIN the default
        directory environment variable. This can be useful in cases where:

            $HIPSCAT_ALMANAC_DIR=$HIPSCAT_DEFAULT_DIR/almanacs/
        """
        default_dir = os.environ.get("HIPSCAT_ALMANAC_DIR", "")
        if default_dir:
            default_dir = os.path.expandvars(default_dir)
        return default_dir

    @classmethod
    def from_catalog_dir(
        cls, catalog_base_dir: str, storage_options: Union[Dict[Any, Any], None] = None
    ) -> Self:
        """Create almanac information from the catalog information found at the target directory"""
        catalog_info = catalog_info_factory.from_catalog_dir(
            catalog_base_dir=file_io.get_file_pointer_from_path(catalog_base_dir),
            storage_options=storage_options,
        )
        args = {
            "catalog_path": catalog_base_dir,
            "catalog_name": catalog_info.catalog_name,
            "catalog_type": catalog_info.catalog_type,
            "catalog_info_object": catalog_info,
            "catalog_info": dataclasses.asdict(catalog_info),
        }
        return cls(**args)

    @classmethod
    def from_file(cls, file: str, storage_options: Union[Dict[Any, Any], None] = None) -> Self:
        """Create almanac information from an almanac file."""
        _, fmt = os.path.splitext(file)
        if fmt != ".yml":
            raise ValueError(f"Unsupported file format {fmt}")
        metadata = file_io.file_io.read_yaml(file, storage_options=storage_options)
        return cls(**metadata)

    def write_to_file(
        self,
        directory=None,
        default_dir=True,
        fmt="yml",
        storage_options: Union[Dict[Any, Any], None] = None,
    ):
        """Write the almanac to an almanac file"""
        if default_dir and directory:
            raise ValueError("Use only one of dir and default_dir")

        if default_dir:
            directory = AlmanacInfo.get_default_dir()

        file_path = file_io.append_paths_to_pointer(
            file_io.get_file_pointer_from_path(directory), f"{self.catalog_name}.{fmt}"
        )
        if file_io.does_file_or_directory_exist(file_path, storage_options=storage_options):
            raise ValueError(f"File already exists at path {str(file_path)}")

        args = {
            "catalog_path": self.catalog_path,
            "catalog_name": self.catalog_name,
            "catalog_type": self.catalog_type,
            "creators": self.creators,
            "description": self.description,
            "catalog_info": self.catalog_info,
        }
        if self.primary:
            args["primary"] = self.primary
        if self.join:
            args["join"] = self.join
        if self.version:
            args["version"] = self.version
        if self.deprecated:
            args["deprecated"] = self.deprecated

        if fmt == "yml":
            encoded_string = yaml.dump(args, sort_keys=False)
        else:
            raise ValueError(f"Unsupported file format {fmt}")

        file_io.write_string_to_file(file_path, encoded_string, storage_options=storage_options)



import dataclasses
import os

from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.io import file_io


def test_read_from_file(
    catalog_info_file_cloud, example_cloud_storage_options, assert_catalog_info_matches_dict
):
    cat_info_fp = file_io.get_file_pointer_from_path(catalog_info_file_cloud)
    catalog_info = CatalogInfo.read_from_metadata_file(
        cat_info_fp, storage_options=example_cloud_storage_options
    )
    for column in [
        "catalog_name",
        "catalog_type",
        "total_rows",
        "epoch",
        "ra_column",
        "dec_column",
    ]:
        assert column in dataclasses.asdict(catalog_info)

    catalog_info_json = file_io.file_io.load_json_file(
        catalog_info_file_cloud, storage_options=example_cloud_storage_options
    )
    assert_catalog_info_matches_dict(catalog_info, catalog_info_json)



import os

import pytest

from hipscat.catalog import Catalog
from hipscat.inspection import plot_pixels, plot_points


@pytest.mark.parametrize("projection", ["moll", "gnom", "cart", "orth"])
def test_generate_map(small_sky_dir_cloud, example_cloud_storage_options, projection):
    """Basic test that map data can be generated (does not test that a plot is rendered)"""
    cat = Catalog.read_from_hipscat(small_sky_dir_cloud, storage_options=example_cloud_storage_options)
    plot_pixels(cat, projection=projection, draw_map=False)
    plot_points(cat, projection=projection, draw_map=False)


def test_generate_map_unknown_projection(small_sky_dir_cloud, example_cloud_storage_options):
    """Test for error with unknown projection type"""

    cat = Catalog.read_from_hipscat(small_sky_dir_cloud, storage_options=example_cloud_storage_options)

    with pytest.raises(NotImplementedError):
        plot_pixels(cat, projection=None, draw_map=False)

    with pytest.raises(NotImplementedError):
        plot_pixels(cat, projection="", draw_map=False)

    with pytest.raises(NotImplementedError):
        plot_pixels(cat, projection="projection", draw_map=False)


def test_generate_map_order1(small_sky_order1_dir_cloud, example_cloud_storage_options):
    """Basic test that map data can be generated (does not test that a plot is rendered)"""
    cat = Catalog.read_from_hipscat(small_sky_order1_dir_cloud, storage_options=example_cloud_storage_options)
    plot_pixels(cat, draw_map=False)
    plot_points(cat, draw_map=False)



import os

import pytest

from hipscat.catalog.association_catalog.association_catalog_info import AssociationCatalogInfo
from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo
from hipscat.catalog.dataset.catalog_info_factory import from_catalog_dir
from hipscat.catalog.index.index_catalog_info import IndexCatalogInfo
from hipscat.catalog.margin_cache.margin_cache_catalog_info import MarginCacheCatalogInfo
from hipscat.catalog.source_catalog.source_catalog_info import SourceCatalogInfo


def test_from_catalog_dir_object(small_sky_dir_cloud, example_cloud_storage_options):
    catalog_info = from_catalog_dir(small_sky_dir_cloud, example_cloud_storage_options)
    assert catalog_info.catalog_name == "small_sky"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, CatalogInfo)


def test_from_catalog_dir_source(source_catalog_info_file_cloud, example_cloud_storage_options):
    catalog_info = from_catalog_dir(
        source_catalog_info_file_cloud, storage_options=example_cloud_storage_options
    )
    assert catalog_info.catalog_name == "small_sky_source_catalog"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, SourceCatalogInfo)


def test_from_catalog_dir_margin_cache(margin_cache_catalog_info_file_cloud, example_cloud_storage_options):
    catalog_info = from_catalog_dir(
        margin_cache_catalog_info_file_cloud, storage_options=example_cloud_storage_options
    )
    assert catalog_info.catalog_name == "margin_cache"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, MarginCacheCatalogInfo)


def test_from_catalog_dir_index(index_catalog_info_file_cloud, example_cloud_storage_options):
    catalog_info = from_catalog_dir(
        index_catalog_info_file_cloud, storage_options=example_cloud_storage_options
    )
    assert catalog_info.catalog_name == "index_catalog"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, IndexCatalogInfo)


def test_from_catalog_dir_association(
    association_catalog_path_cloud, association_catalog_info_file_cloud, example_cloud_storage_options
):
    catalog_info = from_catalog_dir(
        association_catalog_path_cloud, storage_options=example_cloud_storage_options
    )
    assert catalog_info.catalog_name == "small_sky_to_small_sky_order1"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, AssociationCatalogInfo)

    catalog_info = from_catalog_dir(
        association_catalog_info_file_cloud, storage_options=example_cloud_storage_options
    )
    assert catalog_info.catalog_name == "small_sky_to_small_sky_order1"
    assert isinstance(catalog_info, BaseCatalogInfo)
    assert isinstance(catalog_info, AssociationCatalogInfo)



"""Catalog data wrappers"""

from .association_catalog.association_catalog import AssociationCatalog
from .catalog import Catalog
from .catalog_type import CatalogType
from .dataset.dataset import Dataset
from .partition_info import PartitionInfo



import pytest

from hipscat.catalog import Catalog
from hipscat.inspection import plot_pixel_list, plot_pixels, plot_points


@pytest.mark.parametrize("projection", ["moll", "gnom", "cart", "orth"])
def test_generate_map(small_sky_dir, projection):
    """Basic test that map data can be generated (does not test that a plot is rendered)"""

    cat = Catalog.read_from_hipscat(small_sky_dir)
    plot_pixels(cat, projection=projection, draw_map=False)
    plot_points(cat, projection=projection, draw_map=False)


def test_generate_map_unknown_projection(small_sky_dir):
    """Test for error with unknown projection type"""

    cat = Catalog.read_from_hipscat(small_sky_dir)
    with pytest.raises(NotImplementedError):
        plot_pixels(cat, projection=None, draw_map=False)

    with pytest.raises(NotImplementedError):
        plot_pixels(cat, projection="", draw_map=False)

    with pytest.raises(NotImplementedError):
        plot_pixels(cat, projection="projection", draw_map=False)


def test_generate_map_order1(small_sky_order1_dir):
    """Basic test that map data can be generated (does not test that a plot is rendered)"""

    cat = Catalog.read_from_hipscat(small_sky_order1_dir)
    plot_pixels(cat, draw_map=False)
    plot_points(cat, draw_map=False)


def test_visualize_in_memory_catalogs(catalog_info, catalog_pixels):
    """Test behavior of visualization methods for non-on-disk catalogs and pixel data."""
    catalog = Catalog(catalog_info, catalog_pixels)
    plot_pixels(catalog, draw_map=False)
    plot_pixel_list(catalog_pixels, plot_title="My special catalog", draw_map=False)

    with pytest.raises(ValueError, match="on disk catalog required"):
        plot_points(catalog, draw_map=False)
